/* etagit.c
 *
 * Copyright (C) 2004 Ted Phelps
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms fo the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the `etagit' program; see the file `COPYING'.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_WINCONFIG_H
#include <winconfig.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <elvin/elvin.h>
#include "ids.h"
#include "hexdump.h"
#define HAVE_UINT64_T 1 /* Guaranteed by <elvin/elvin.h> */
#include "librfid/rfid-interface.h"

#define DEFAULT_POLL_PERIOD 250
#define MODEL_DETECT 0
#define MODEL_S6000 6000
#define MODEL_S6500 6500

/* The amount of time to wait for the tag reader to start sending its
 * response after a request was sent */
#define MAJOR_TIMEOUT 1000

/* The amount of time to wait for more data after some of a response
 * was read */
#define MINOR_TIMEOUT 50

typedef enum
{
    STATE_START,
    STATE_VERS_PENDING,
    STATE_INV_PENDING,
    STATE_PAUSE,
    STATE_RESET
} reader_state_t;

static char *state_names[] =
{
    "START",
    "VERS_PENDING",
    "INV_PENDING",
    "PAUSE",
    "RESET"
};

typedef SERIALPORT_T serialport_t;

/* The following table contains both Unix- and Windows-style baud rate
 * declarations.  This will have to be rethought if we ever encounter
 * a system which supports both simultaneously. */
static int baud_table[][2] =
{
#ifdef B50
    { 50, B50 },
#endif
#ifdef B75
    { 75, B75 },
#endif
#ifdef B110
    { 110, B110 },
#endif
#ifdef CBR_110
    { 110, CBR_110 },
#endif
#ifdef B134
    { 134, B134 },
#endif
#ifdef B150
    { 150, B150 },
#endif
#ifdef B200
    { 200, B200 },
#endif
#ifdef B300
    { 300, B300 },
#endif
#ifdef CBR_300
    { 300, CBR_300 },
#endif
#ifdef B600
    { 600, B600 },
#endif
#ifdef CBR_600
    { 600, CBR_600 },
#endif
#ifdef B1200
    { 1200, B1200 },
#endif
#ifdef CBR_1200
    { 1200, CBR_1200 },
#endif
#ifdef B1800
    { 1800, B1800 },
#endif
#ifdef B2400
    { 2400, B2400 },
#endif
#ifdef CBR_2400
    { 2400, CBR_2400 },
#endif
#ifdef B4800
    { 4800, B4800 },
#endif
#ifdef CBR_4800
    { 4800, CBR_4800 },
#endif
#ifdef B9600
    { 9600, B9600 },
#endif
#ifdef CBR_9600
    { 9600, CBR_9600 },
#endif
#ifdef CBR_14400
    { 14400, CBR_14400 },
#endif
#ifdef B19200
    { 19200, B19200 },
#endif
#ifdef CBR_19200
    { 19200, CBR_19200 },
#endif
#ifdef B38400
    { 38400, B38400 },
#endif
#ifdef CBR_38400
    { 38400, CBR_38400 },
#endif
#ifdef B57600
    { 57600, B57600 },
#endif
#ifdef CBR_56000
    { 56000, CBR_56000 },
#endif
#ifdef CBR_57600
    { 57600, CBR_57600 },
#endif
#ifdef B76800
    { 76800, B76800 },
#endif
#ifdef B115200
    { 115200, B115200 },
#endif
#ifdef CBR_115200
    { 115200, CBR_115200 },
#endif
#ifdef CBR_128000
    { 128000, CBR_128000 },
#endif
#ifdef B153600
    { 153600, B153600 },
#endif
#ifdef B230400
    { 230400, B230400 },
#endif
#ifdef CBR_256000
    { 256000, CBR_256000 },
#endif
#ifdef B307200
    { 307200, B307200 },
#endif
#ifdef B460800
    { 460800, B460800 },
#endif
#ifdef B500000
    { 500000, B500000 },
#endif
#ifdef B576000
    { 576000, B576000 },
#endif
#ifdef B921600
    { 921600, B921600 },
#endif
#ifdef B1000000
    { 1000000, B1000000 },
#endif
#ifdef B1152000
    { 1152000, B1152000 },
#endif
#ifdef B2000000
    { 2000000, B2000000 },
#endif
#ifdef B2500000
    { 2500000, B2500000 },
#endif
#ifdef B3000000
    { 3000000, B3000000 },
#endif
#ifdef B3500000
    { 3500000, B3500000 },
#endif
#ifdef B4000000
    { 4000000, B4000000 },
#endif

    { -1, -1 }
};

/* Global variables */
static char *progname = "etagit";
static elvin_client_t client = NULL;
static elvin_error_t error = NULL;
static elvin_handle_t handle = NULL;
static serialport_t port;
static int rfid_model = MODEL_DETECT;
static int poll_period = DEFAULT_POLL_PERIOD;
static int verbosity = 0;
static int baud_count;
static struct rfid_tag *tags;
static size_t tag_count;
static rfid_interface_t ifaces[2];
static int iface_count = 0;
static int iface_index = 0;
static reader_state_t rfid_state = STATE_START;
static elvin_timeout_t response_timeout = NULL;
static unsigned long long last_ids[RFID_MAX_IDS];
static int last_id_count = 0;
#ifdef WIN32
DWORD serial_status;
static elvin_win32_event_handler_t event_handler = NULL;
static OVERLAPPED overlapped;
#endif /* WIN32 */

#define OPTIONS "a:e:S:d:hp:s:Vvw"
static void usage(void)
{
    fprintf(stderr, "usage: %s [OPTION]... tag-file\n"
            "  " OPTION_PREFIX "e elvin-url\tURL of an Elvin router\n"
            "  " OPTION_PREFIX "S scope\tScope of an Elvin router\n"
            "  " OPTION_PREFIX "d device\tSerial port to use [" DEFAULT_DEVICE "]\n"
            "  " OPTION_PREFIX "s model\tTag reader model (6000 or 6500) [auto-detected]\n"
            "  " OPTION_PREFIX "p pause\tAmount of time to pause between polls of RF-ID reader [%d]\n"
            "  " OPTION_PREFIX "v\t\tIncrease output verbosity\n"
            "  " OPTION_PREFIX "h\t\tPrint this brief help summary\n"
            "  " OPTION_PREFIX "V\t\tPrint version information and exit\n",
            progname, DEFAULT_POLL_PERIOD);
}

static void dprintf(int level, const char *format, ...)
{
    va_list ap;

    /* Don't print if the verbosity level is too low */
    if (verbosity < level) {
        return;
    }

    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
}

#ifdef WIN32
static int open_serial_port(char *path, serialport_t *port_out)
{
    serialport_t port;
    char *buffer;
    int length;

    /* Measure the path */
    length = strlen(path);

    /* Allocate enough memory for the path */
    if ((buffer = malloc(length + 2)) == NULL) {
        return GetLastError();
    }
    sprintf(buffer, "%s:", path);

    if ((port = CreateFile(buffer,
                           GENERIC_READ | GENERIC_WRITE, 0,
                           NULL, OPEN_EXISTING,
                           FILE_FLAG_OVERLAPPED,
                           NULL)) == INVALID_HANDLE_VALUE) {
        free(buffer);
        return GetLastError();
    }

    *port_out = port;
    free(buffer);
    return 0;
}
#else /* !WIN32 */
static int open_serial_port(char *path, serialport_t *port_out)
{
    serialport_t port;

    if ((port = open(path, O_RDWR | O_NOCTTY)) < 0) {
        return errno;
    }

    *port_out = port;
    return 0;
}
#endif /* WIN32 */

/* Find an entry in the baud table */
static int compare_baud(const void *key, const void *entry)
{
    int keyval = *(int *)key;
    int entval = *(int *)entry;

    return keyval < entval ? -1 : keyval > entval ? 1 : 0;
}

/* Compares two RF-ID tag ids */
static int compare_ids(const void *id1, const void *id2)
{
    unsigned long long v1 = *(unsigned long long *)id1;
    unsigned long long v2 = *(unsigned long long *)id2;

    return v1 < v2 ? -1 : v1 > v2 ? 1 : 0;
}

#ifdef WIN32
static int config_serial_port(serialport_t port,
                              int baud,
                              int data_bits,
                              char parity,
                              int stop_bits)
{
    DCB dcb;
    COMMTIMEOUTS timeouts;
    int *entry;

    dprintf(5, "%s: configuring serial port for %d %d%c%d\n",
            progname, baud, data_bits, parity, stop_bits);

    /* Get the current settings */
    dcb.DCBlength = sizeof(dcb);
    if (!GetCommState(port, &dcb)) {
        fprintf(stderr, "%s: error: GetCommState failed: %d\n", progname, GetLastError());
        exit(1);
    }

    /* Look up the baud rate in our table */
    if ((entry = bsearch(&baud,
                         baud_table, baud_count,
                         2 * sizeof(int),
                         compare_baud)) == NULL) {
        fprintf(stderr, "%s: error: unsupported baud rate: %d\n", progname, baud);
        exit(1);
    }

    dcb.BaudRate = entry[1];
    dcb.fBinary = TRUE;
    dcb.fAbortOnError = FALSE;
    dcb.fNull = FALSE;
    dcb.ByteSize = data_bits;

    switch (parity)
    {
        case 'n':
        case 'N':
            dcb.fParity = FALSE;
            dcb.Parity = NOPARITY;
            break;

        case 'e':
        case 'E':
            dcb.fParity = TRUE;
            dcb.Parity = EVENPARITY;
            break;

        case 'o':
        case 'O':
            dcb.fParity = TRUE;
            dcb.Parity = ODDPARITY;
            break;

        default:
            fprintf(stderr, "%s: error: unsupported parity setting: %c\n", progname, parity);
            exit(1);
    }

    if (stop_bits == 1) {
        dcb.StopBits = ONESTOPBIT;
    } else if (stop_bits == 15) {
        dcb.StopBits = ONE5STOPBITS;
    } else if (stop_bits == 2) {
        dcb.StopBits = TWOSTOPBITS;
    } else {
        fprintf(stderr, "%s: error: unsupported stop bits: %d\n", progname, stop_bits);
        exit(1);
    }

    if (!SetCommState(port, &dcb)) {
        fprintf(stderr, "%s: error: SetCommState failed: %s\n",
                progname, strerror(GetLastError()));
        exit(1);
    }

    /* Set the timeouts */
    memset(&timeouts, 0, sizeof(COMMTIMEOUTS));
    timeouts.ReadIntervalTimeout = MAXDWORD;
    timeouts.ReadTotalTimeoutMultiplier = MAXDWORD;
    timeouts.ReadTotalTimeoutConstant = 1000;
    if (!SetCommTimeouts(port, &timeouts)) {
        fprintf(stderr, "%s: error: SetCommTimeouts failed: %s\n",
                progname, strerror(GetLastError()));
        exit(1);
    }

    /* Flush the buffer */
    if (!PurgeComm(port,
                   PURGE_TXABORT | PURGE_RXABORT |
                   PURGE_TXCLEAR | PURGE_RXCLEAR)) {
        fprintf(stderr, "%s: error: PurgeComm failed: %s\n",
               progname, strerror(GetLastError()));
        exit(1);
    }

    return 0;
}
#else /* !WIN32 */
static int config_serial_port(serialport_t port,
                              int baud,
                              int data_bits,
                              char parity,
                              int stop_bits)
{
    struct termios termios;
    int *entry;

    dprintf(5, "%s: configuring serial port for %d %d%c%d\n",
            progname, baud, data_bits, parity, stop_bits);

    memset(&termios, 0, sizeof(struct termios));

    /* CLOCAL: ignore the modem control lines,
     * CREAD: enable the receiver */
    termios.c_cflag = CLOCAL | CREAD;

    /* Look up the baud setting */
    if ((entry = bsearch(&baud,
                         baud_table, baud_count,
                         2 * sizeof(int),
                         compare_baud)) == NULL) {
        fprintf(stderr, "%s: error: unsupported baud rate: %d\n", progname, baud);
        exit(1);
    }
    termios.c_cflag |= entry[1];

    /* Configure the number of data bits */
    switch (data_bits)
    {
        case 5:
            termios.c_cflag |= CS5;
            break;

        case 6:
            termios.c_cflag |= CS6;
            break;

        case 7:
            termios.c_cflag |= CS7;
            break;

        case 8:
            termios.c_cflag |= CS8;
            break;

        default:
            fprintf(stderr, "%s: error: unsupported data bit count: %d\n",
                    progname, data_bits);
            exit(1);
    }

    /* Configure the parity */
    switch (parity)
    {
        case 'n':
        case 'N':
            /* This is the default */
            break;

        case 'e':
        case 'E':
            /* Enable parity, default is even */
            termios.c_cflag |= PARENB;
            break;

        case 'o':
        case 'O':
            termios.c_cflag |= PARENB | PARODD;
            break;

        default:
            fprintf(stderr, "%s: error: unsupported parity setting: %c\n", progname, parity);
            exit(1);
    }

    /* Configure the stop bits */
    if (stop_bits == 1) {
        /* One stop bit is the default */
    } else if (stop_bits == 2) {
        termios.c_cflag |= CSTOPB;
    } else {
        fprintf(stderr, "%s: error: unsupported number of stop bits: %d\n",
                progname, stop_bits);
    }

    /* Clear the line */
    if (tcflush(port, TCIFLUSH) < 0) {
        fprintf(stderr, "%s: error: tcflush failed: %s\n",
                progname, strerror(errno));
        exit(1);
    }

    /* Activate our settings */
    if (tcsetattr(port, TCSANOW, &termios) < 0) {
        fprintf(stderr, "%s: error: tcsetattr failed: %s\n",
                progname, strerror(errno));
        exit(1);
    }

    return 0;
}
#endif /* WIN32 */

/* Close the serial port */
static int close_serial_port(serialport_t port)
{
#ifdef WIN32
    if (!CloseHandle(port)) {
        return GetLastError();
    }
#else /* !WIN32 */
    if (close(port) < 0) {
        return errno;
    }
#endif /* WIN32 */

    return 0;
}

/* Forward declarations for the state machine */

static int ELVIN_CALLBACK response_timeout_cb(elvin_timeout_t timeout,
                                              void *rock,
                                              elvin_error_t error);

#ifdef WIN32
static int serial_read_cb(elvin_win32_event_handler_t handler,
                          HANDLE event,
                          void *rock,
                          elvin_error_t error);
#else /* !WIN32 */
static int ELVIN_CALLBACK serial_read_cb(elvin_io_handler_t handler,
                                         elvin_socket_t fd,
                                         uint32_t active_mask,
                                         int *would_block_out,
                                         void *rock,
                                         elvin_error_t error);
#endif /* WIN32 */

/* Sends a request to the serial port */
static int send_request(int type)
{
    int err;
    char buffer[256];
    int count;
#ifdef WIN32
    DWORD length;
#else /* !WIN32 */
    char *point;
    char *end;
    ssize_t length;
#endif /* WIN32 */

    /* Sanity check */
    if (response_timeout != NULL) {
        fprintf(stderr, "%s: internal error: sending request at inappropriate time\n",
                progname);
        exit(1);
    }

    /* Create the requst */
    count = 256;
    if ((err = rfid_interface_make_req(ifaces[iface_index], type, buffer, &count)) != 0) {
        return err;
    }

#ifdef WIN32
    /* Write the request to the serial port */
    if (!WriteFile(port, buffer, count, NULL, &overlapped)) {
        if ((err = GetLastError()) != ERROR_IO_PENDING) {
            fprintf(stderr, "%s: error: WriteFile failed: %d\n", progname, err);
            exit(1);
        }
    }

    /* Wait for the operation to complete */
    if (!GetOverlappedResult(port, &overlapped, &length, TRUE)) {
        fprintf(stderr, "%s: error: GetOverlappedResult failed: %d\n", progname, GetLastError());
        exit(1);
    }

    /* Sanity check */
    if (count != length) {
        fprintf(stderr, "%s: error: incomplete write", progname);
        exit(1);
    }

    dprintf(3, "%s: wrote %d-byte frame\n", progname, count);
    if (verbosity >= 3) {
        fhexdump(buffer, count, stderr);
    }

    /* Wait for the serial port to become readable */
    if (!SetCommMask(port, EV_BREAK | EV_ERR | EV_RXCHAR)) {
        fprintf(stderr, "%s: SetCommMask failed: %d\n", progname, GetLastError());
        exit(1);
    }

    if (!WaitCommEvent(port, &serial_status, &overlapped)) {
        if ((err = GetLastError()) != ERROR_IO_PENDING) {
            fprintf(stderr, "%s: error: WaitCommEvent failed: %d\n", progname, err);
            exit(1);
        }

        if ((event_handler = elvin_win32_add_event_handler(client,
                                                           overlapped.hEvent,
                                                           serial_read_cb, NULL,
                                                           error)) == NULL) {
            fprintf(stderr, "%s: error: ", progname);
            elvin_error_fprintf(stderr, error);
            exit(1);
        }
    }
#else /* !WIN32 */
    /* Write the request to the serial port */
    point = buffer;
    end = buffer + count;

    while (point < end)
    {
        dprintf(3, "%s: writing frame: %d bytes left\n", progname, end - point);

        if ((length = write(port, point, end - point)) < 0) {
            fprintf(stderr, "%s: write failed: %s\n",
                    progname, strerror(errno));
            exit(1);
        }

        if (verbosity >= 3) {
            fhexdump(point, length, stderr);
        }

        point += length;
    }
#endif /* WIN32 */

    /* Give the request a little while to be processed */
    if ((response_timeout = elvin_mainloop_add_timeout(client,
                                                       MAJOR_TIMEOUT,
                                                       response_timeout_cb, NULL,
                                                       error)) == NULL) {
        fprintf(stderr, "%s: error: ", progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    return 0;
}


/* Change to the given state */
static int switch_to_state(reader_state_t new_state)
{
    int err;
#ifdef WIN32
    DWORD result;
#endif /* WIN32 */

    dprintf(4, "%s: %s->%s\n",
            progname,
            state_names[rfid_state],
            state_names[new_state]);

    switch (new_state)
    {
        case STATE_START:
            fprintf(stderr, "%s: error: transition to start state\n", progname);
            exit(1);

        case STATE_VERS_PENDING:
            /* Sanity check the transition */
            if (rfid_state != STATE_START) {
                fprintf(stderr, "%s: error: can't go from %s to %s\n",
                        progname,
                        state_names[rfid_state],
                        state_names[new_state]);
                exit(1);
            }

            /* Send a version request */
            if ((err = send_request(RFID_TYPE_VERSION)) != 0) {
                dprintf(0, "%s: warning: send_request failed: %d\n",
                        progname, err);
                return switch_to_state(STATE_RESET);
            }

            break;

        case STATE_INV_PENDING:
            /* Sanity check the transition */
            if (rfid_state != STATE_PAUSE) {
                fprintf(stderr, "%s: error: can't go from %s to %s\n",
                        progname,
                        state_names[rfid_state],
                        state_names[new_state]);
                exit(1);
            }

            /* Send an inventory request */
            if ((err = send_request(RFID_TYPE_INVENTORY)) != 0) {
                dprintf(0, "%s: warning: send_inventory_request failed: %d\n",
                        progname, err);
                return switch_to_state(STATE_RESET);
            }

            break;

        case STATE_PAUSE:
            /* Sanity check the transition */
            if (rfid_state != STATE_VERS_PENDING &&
                rfid_state != STATE_INV_PENDING) {
                fprintf(stderr, "%s: error: can't go from %s to %s\n",
                        progname,
                        state_names[rfid_state],
                        state_names[new_state]);
                exit(1);
            }

            /* Sleep for a little while */
            if ((response_timeout = elvin_mainloop_add_timeout(client,
                                                               poll_period,
                                                               response_timeout_cb, NULL,
                                                               error)) == NULL) {
                fprintf(stderr, "%s: error: ", progname);
                elvin_error_fprintf(stderr, error);
                exit(1);
            }
            break;

        case STATE_RESET:
            /* Cancel pending timeouts */
            if (response_timeout != NULL) {
                if (!elvin_mainloop_remove_timeout(client, response_timeout, error)) {
                    fprintf(stderr, "%s: error: ", progname);
                    elvin_error_fprintf(stderr, error);
                    exit(1);
                }
            }

#ifdef WIN32
            /* If a WaitCommEvent was pending then kill it */
            if (event_handler != NULL) {
                /* Cancel any pending I/O operations */
                if (!CancelIo(port)) {
                    fprintf(stderr, "%s: error: CancelIo failed: %d\n", progname, GetLastError());
                    exit(1);
                }

                /* Collect the broken remains of WaitCommEvent */
                if (!GetOverlappedResult(port, &overlapped, &result, FALSE)) {
                    if ((err = GetLastError()) != ERROR_OPERATION_ABORTED) {
                        fprintf(stderr, "%s: error: GetOverlappedResult failed: %d\n",
                                progname, GetLastError());
                        exit(1);
                    }
                }

                /* Kill the event handler */
                if (!elvin_win32_remove_event_handler(client, event_handler, error)) {
                    fprintf(stderr, "%s: error: ", progname);
                    elvin_error_fprintf(stderr, error);
                    exit(1);
                }

                event_handler = NULL;
            }

            /* Purge the serial port */
            if (!PurgeComm(port,
                           PURGE_TXABORT | PURGE_RXABORT |
                           PURGE_TXCLEAR | PURGE_RXCLEAR)) {
                fprintf(stderr, "%s: error: PurgeComm failed: %d\n",
                        progname, GetLastError());
                exit(1);
            }
#else /* !WIN32 */
            if (tcflush(port, TCIOFLUSH) < 0) {
                fprintf(stderr, "%s: error: tcflush failed: %s\n",
                        progname, strerror(errno));
                exit(1);
            }
#endif /* WIN32 */

            /* Reset the interface */
            rfid_interface_reset(ifaces[iface_index]);

            /* Start over from scratch */
            rfid_state = STATE_START;
            if ((err = switch_to_state(STATE_VERS_PENDING)) != 0) {
                fprintf(stderr, "%s: error: switch_to_state failed: %d\n", progname, err);
                exit(1);
            }

            return 0;

        default:
            fprintf(stderr, "%s: error: invalid reader state: %d\n",
                    progname, new_state);
            exit(1);
    }

    rfid_state = new_state;
    return 0;
}

/* Called when our attempt to connect to an Elvin server has finished */
static int ELVIN_CALLBACK connect_cb(elvin_handle_t handle,
                                     int result,
                                     void *rock,
                                     elvin_error_t error)
{
    int err;

    /* Complain if we failed to connect */
    if (!result) {
        dprintf(1, "  failed\n");

        elvin_error_fprintf(stderr, error);
        if (!elvin_mainloop_quit(client, error)) {
            fprintf(stderr, "%s: elvin_mainloop_quit failed\n%s: error: ",
                    progname, progname);
            elvin_error_fprintf(stderr, error);
            exit(1);
        }

        return 0;
    }

    dprintf(1, "  done\n");

#ifndef WIN32
    /* Listen for serial port activity */
    if (!elvin_mainloop_add_io_handler(client, port,
                                       ELVIN_READ_MASK,
                                       serial_read_cb, NULL,
                                       error)) {
        fprintf(stderr, "%s: error: ", progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }
#endif /* !WIN32 */

    /* Transition to the version pending state by sending out a
     * version request */
    if ((err = switch_to_state(STATE_VERS_PENDING)) != 0) {
        fprintf(stderr, "%s: error: switch_to_state failed: %d\n", progname, err);
        exit(1);
    }

    return 1;
}

/* Send a notification about a tag */
static int send_notification(unsigned long long id, int was_added)
{
    elvin_notification_t notification;
    char *action = was_added ? "added" : "removed";
    char *name;

    /* Look up the tag name */
    if ((name = find_id(tags, tag_count, id)) == NULL) {
        name = "";
    }

    /* Print debugging info */
    dprintf(1, "%s: id=%" ELVIN_UINT64_PRINT " name=\"%s\" %s\n",
            progname, id, name, action);

    
    /* Create a notification */
    if ((notification = elvin_notification_alloc(client, error)) == NULL) {
        fprintf(stderr, "%s: ", progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    /* Add some fields to it */
    if (!elvin_notification_add_string(notification, "User-Agent", PACKAGE "-" VERSION, error) ||
        !elvin_notification_add_string(notification, "Action", action, error) ||
        !elvin_notification_add_int64(notification, "Tag-Id", id, error) ||
        !elvin_notification_add_string(notification, "Name", name, error)) {
        fprintf(stderr, "%s: ", progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    /* Send it */
    if (!elvin_async_notify(handle, notification, 1, NULL, error)) {
        fprintf(stderr, "%s: ", progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    return 0;
}

/* Called when a request is ignored */
static int ELVIN_CALLBACK response_timeout_cb(elvin_timeout_t timeout,
                                              void *rock,
                                              elvin_error_t error)
{
    int err, baud, data_bits, stop_bits;
    char parity;

    /* Clear the timeout */
    response_timeout = NULL;

    switch (rfid_state)
    {
        case STATE_VERS_PENDING:
            /* We didn't get a response to our version request, so try
             * a different reader protocol */
            iface_index = (iface_index + 1) % iface_count;

            /* Configure the serial port for this type of tag reader */
            rfid_interface_get_settings(ifaces[iface_index], &baud, &data_bits, &parity, &stop_bits);
            if ((err = config_serial_port(port, baud, data_bits, parity, stop_bits)) != 0) {
                fprintf(stderr, "%s: unable to configure serial port: %d\n",
                        progname, err);
                exit(1);
            }

            break;

        case STATE_INV_PENDING:
            /* No reply to our inventory request. */
            dprintf(1, "%s: Lost connection to tag reader: operation timed out\n", progname);
            break;

        case STATE_PAUSE:
            /* That was just our pause, so do the inventory thing */
            if ((err = switch_to_state(STATE_INV_PENDING)) != 0) {
                fprintf(stderr, "%s: error: switch_to_state failed\n", progname);
                exit(1);
            }

            return 1;

        case STATE_START:
        case STATE_RESET:
            fprintf(stderr, "%s: warning: operation timed out in state %s\n",
                    progname, state_names[rfid_state]);

        default:
            fprintf(stderr, "%s: error: reader in unknown state: %d\n",
                    progname, rfid_state);
            exit(1);
    }

    /* Reset the reader and try again */
    if ((err = switch_to_state(STATE_RESET)) != 0) {
        fprintf(stderr, "%s: switch_to_state failed: %d\n", progname, err);
        exit(1);
    }

    return 1;
}

static int update_ids(unsigned long long *ids, int count)
{
    int err;
    int old_next, old_index;
    int new_next, new_index;

    /* Sort the ids */
    qsort(ids, count, sizeof(unsigned long long), compare_ids);

    /* Determine which ids have gone and which are new */
    old_next = old_index = new_next = new_index = 0;
    while (old_index < last_id_count && new_index < count) {
        if (last_ids[old_index] < ids[new_index]) {
            if ((err = send_notification(last_ids[old_index++], 0)) != 0) {
                return err;
            }
        } else if (last_ids[old_index] > ids[new_index]) {
            if ((err = send_notification(ids[new_index++], 1)) != 0) {
                return err;
            }
        } else {
            /* Unchanged */
            old_index++;
            new_index++;
        }
    }

    /* Deal with the left over ids */
    while (old_index < last_id_count) {
        if ((err = send_notification(last_ids[old_index++], 0)) != 0) {
            return err;
        }
    }

    while (new_index < count) {
        if ((err = send_notification(ids[new_index++], 1)) != 0) {
            return err;
        }
    }

    /* Record the ids */
    memcpy(last_ids, ids, count * sizeof(unsigned long long));
    last_id_count = count;
    return 0;
}

static int digest_input(char *buffer, int count)
{
    int err;
    int is_complete;
    union rfid_response response;
    int i;

    dprintf(3, "%s: read %d bytes\n", progname, count);
    if (verbosity >= 3) {
        fhexdump(buffer, count, stderr);
    }

    /* Feed those bytes to the interface */
    if ((err = rfid_interface_feed(ifaces[iface_index],
                                   buffer, count,
                                   &is_complete, &response)) != 0) {
    }

    /* If there's more to come then set a shorter timeout */
    if (!is_complete) {
        return 0;
    }

    dprintf(3, "%s: complete frame received\n", progname);
    switch (response.type)
    {
        case RFID_TYPE_VERSION:
            dprintf(1, "%s: Tag reader found: %s\n",
                    progname,
                    response.version.string);
            if ((err = switch_to_state(STATE_PAUSE)) != 0) {
                fprintf(stderr, "%s: error: switch_to_state failed: %d\n", progname, err);
                exit(1);
            }

            break;

        case RFID_TYPE_INVENTORY:
            dprintf(5, "%s:   inventory (%d tags)\n",
                    progname,
                    response.inventory.count);
            for (i = 0; i < response.inventory.count; i++) {
                dprintf(5, "%s:     %lld\n",
                        progname,
                        response.inventory.ids[i]);
            }

            /* Update the state and send out notifications */
            if ((err = update_ids(response.inventory.ids, response.inventory.count)) != 0) {
                fprintf(stderr, "%s: error: update_ids failed: %d\n", progname, err);
                exit(1);
            }

            /* Go to the pause state */
            if ((err = switch_to_state(STATE_PAUSE)) != 0) {
                fprintf(stderr, "%s: error: switch_to_state failed %d\n", progname, err);
                exit(1);
            }
            break;

        case RFID_TYPE_RESET:
            dprintf(5, "%s:   reset\n", progname);
            break;

        default:
            fprintf(stderr, "%s: error: unsupported response type: %d\n",
                    progname,
                    response.type);
            exit(1);
    }

    return 1;
}

#ifdef WIN32
static int serial_read_cb(elvin_win32_event_handler_t handler,
                          HANDLE event,
                          void *rock,
                          elvin_error_t error)
{
    int err;
    DWORD count, status;
    char buffer[256];


    /* Cancel the I/O handler */
    if (event_handler != NULL) {
        if (!elvin_win32_remove_event_handler(client, event_handler, error)) {
            fprintf(stderr, "%s: error: ", progname);
            elvin_error_fprintf(stderr, error);
            exit(1);
        }

        event_handler = NULL;
    }

    /* Wait for the operation to complete */
    if (!GetOverlappedResult(port, &overlapped, &count, TRUE)) {
        fprintf(stderr, "%s: error: GetOverlappedResult failed: %d\n", progname, GetLastError());
        exit(1);
    }

    status = serial_status;
    serial_status = 0;

    /* This shouldn't happen anymore */
    if (status == 0) {
        return 1;
    }

    /* Cancel the response timeout */
    if (response_timeout != NULL) {
        if (!elvin_mainloop_remove_timeout(client, response_timeout, error)) {
            fprintf(stderr, "%s: error: ", progname);
            elvin_error_fprintf(stderr, error);
            exit(1);
        }

        response_timeout = NULL;
    }

    /* Clear an error condition */
    if (status & EV_ERR) {
        dprintf(3, "%s: comm ERR\n", progname);
        if (!ClearCommError(port, &count, NULL)) {
            fprintf(stderr, "%s: ClearCommError failed: %d\n", progname, GetLastError());
            exit(1);
        }
    }

    /* Clear a break condition */
    if (status & EV_BREAK) {
        dprintf(3, "%s: BREAK received\n", progname);
        if (!ClearCommBreak(port)) {
            fprintf(stderr, "%s: ClearCommBreak failed: %d\n", progname, GetLastError());
            exit(1);
        }
    }

    /* Data awaits? */
    if (status & EV_RXCHAR) {
        /* Read it */
        if (!ReadFile(port, buffer, 256, NULL, &overlapped)) {
            if ((err = GetLastError()) != ERROR_IO_PENDING) {
                fprintf(stderr, "%s: error: ReadFile failed: %d\n", progname, err);
                exit(1);
            }
        }

        /* Wait for the read to complete */
        if (!GetOverlappedResult(port, &overlapped, &count, TRUE)) {
            fprintf(stderr, "%s: error: GetOverlappedResult failed: %d\n", progname, GetLastError());
            exit(1);
        }

        /* Digest what we've read */
        if (digest_input(buffer, count)) {
            return 1;
        }
    }

    /* Response not complete; set a shorter timeout. */
    if ((response_timeout = elvin_mainloop_add_timeout(client,
                                                       MINOR_TIMEOUT,
                                                       response_timeout_cb, NULL,
                                                       error)) == NULL) {
        fprintf(stderr, "%s: error: ", progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    /* We're interested in breaks, errors and data available */
    if (!SetCommMask(port, EV_BREAK | EV_ERR | EV_RXCHAR)) {
        fprintf(stderr, "%s: SetCommMask failed: %d\n", progname, GetLastError());
        exit(1);
    }

    /* Wait for something to happen */
    if (!WaitCommEvent(port, &serial_status, &overlapped)) {
        if ((err = GetLastError()) != ERROR_IO_PENDING) {
            fprintf(stderr, "%s: error: WaitCommEvent failed: %d\n", progname, err);
            exit(1);
        }

        if ((event_handler = elvin_win32_add_event_handler(client,
                                                           overlapped.hEvent,
                                                           serial_read_cb, NULL,
                                                           error)) == NULL) {
            fprintf(stderr, "%s: error: ", progname);
            elvin_error_fprintf(stderr, error);
            exit(1);
        }
    }

    return 1;
}
#else /* !WIN32 */
static int ELVIN_CALLBACK serial_read_cb(elvin_io_handler_t handler,
                                         elvin_socket_t fd,
                                         uint32_t active_mask,
                                         int *would_block_out,
                                         void *rock,
                                         elvin_error_t error)
{
    char buffer[256];
    int length;

    /* Cancel the response timeout */
    if (response_timeout != NULL) {
        if (!elvin_mainloop_remove_timeout(client, response_timeout, error)) {
            fprintf(stderr, "%s: error: ", progname);
            elvin_error_fprintf(stderr, error);
            exit(1);
        }

        response_timeout = NULL;
    }

    /* Read what we can */
    if ((length = read(port, buffer, 256)) < 0) {
        fprintf(stderr, "%s: error: read failed: %s\n",
                progname, strerror(errno));
        exit(1);
    }

    if (digest_input(buffer, length)) {
        return 1;
    }

    /* We don't have a complete packet yet; set a shorter timeout. */
    if ((response_timeout = elvin_mainloop_add_timeout(client,
                                                       MINOR_TIMEOUT,
                                                       response_timeout_cb, NULL,
                                                       error)) == NULL) {
        fprintf(stderr, "%s: error: ", progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    return 1;
}
#endif

int main(int argc, char *argv[])
{
    int err;
    int choice;
    char *devname = DEFAULT_DEVICE;
    char *point;
    FILE *file;
    int baud, data_bits, stop_bits;
    char parity;

    /* Strip leading directories from the executable name */
    if ((progname = strrchr(argv[0], DIR_SEPARATOR)) == NULL) {
        progname = argv[0];
    } else {
        progname++;
    }

    /* Count the number of entries in baud_table */
    for (baud_count = 0; baud_table[baud_count][0] > 0; baud_count++);

    /* Get an Elvin error context */
    if ((error = elvin_error_alloc(NULL, NULL)) == NULL) {
        fprintf(stderr, "%s: error: elvin_error_alloc failed\n", progname);
        exit(1);
    }

#ifdef WIN32
    /* Initialize the Elvin client library */
    if ((client = elvin_win32_init_default(error)) == NULL) {
        fprintf(stderr, "%s: error: elvin_win32_init_default failed:\n%s: error: ",
                progname, progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }
#else /* !WIN32 */
    /* Initialize the Elvin client library */
    if ((client = elvin_default_init_default(error)) == NULL) {
        fprintf(stderr, "%s: error: elvin_default_init_default failed:\n%s: error: ",
                progname, progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }
#endif /* WIN32 */

    /* Allocate a connection handle */
    if ((handle = elvin_handle_alloc(client, error)) == NULL) {
        fprintf(stderr, "%s: error: elvin_handle_alloc failed:\n%s: error: ",
                progname, progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    /* Parse the command-line options */
    while ((choice = getopt(argc, argv, OPTIONS)) != EOF)
    {
        switch (choice)
        {
            case 'a':
                /* Antenna count */
                fprintf(stderr, "%s: error: antenna hack no longer supported\n", progname);
                exit(1);

            case 'e':
                /* Elvin URL */
                if (!elvin_handle_append_url(handle, optarg, error)) {
                    fprintf(stderr, "%s: error: ", progname);
                    elvin_error_fprintf(stderr, error);
                    exit(1);
                }

                break;

            case 'S':
                /* Elvin discovery scope */
                if (!elvin_handle_set_discovery_scope(handle, optarg, error)) {
                    fprintf(stderr, "%s: error: ", progname);
                    elvin_error_fprintf(stderr, error);
                    exit(1);
                }

                break;

            case 'd':
                /* Device name */
                devname = optarg;
                break;

            case 'p':
                /* Poll period */
                poll_period = strtol(optarg, &point, 0);
                if (point == optarg || *point != 0 || poll_period < 0) {
                    fprintf(stderr, "%s: error: invalid poll period: %s\n", progname, optarg);
                    exit(1);
                }

                break;

            case 's':
                /* Reader model */
                if (strcmp(optarg, "6000") == 0) {
                    rfid_model = MODEL_S6000;
                } else if (strcmp(optarg, "6500") == 0) {
                    rfid_model = MODEL_S6500;
                } else {
                    fprintf(stderr, "%s: error: invalid RF-ID reader model: %s\n",
                            progname, optarg);
                    exit(1);
                }

                break;

            case 'w':
                /* Wood hack */
                fprintf(stderr, "%s: error: wood hack no longer supported\n", progname);
                exit(1);

            case 'h':
                usage();
                exit(0);

            case 'V':
                fprintf(stderr, "%s (" PACKAGE ") version " VERSION "\n", progname);
                exit(0);

            case 'v':
                verbosity++;
                break;

            case '?':
                usage();
                exit(1);

            default:
                fprintf(stderr, "%s: internal error: unhandled option -%c\n", progname, choice);
                exit(1);
        }
    }

    /* Allocate the rfid interfaces */
    if (rfid_model == MODEL_DETECT || rfid_model == MODEL_S6000) {
        if ((err = rfid_interface_alloc(&s6000_funcs, &ifaces[iface_count++])) != 0) {
            fprintf(stderr, "%s: error: unable to allocate rfid interface: %s\n",
                    progname, strerror(err));
            exit(1);
        }
    } 

    if (rfid_model == MODEL_DETECT || rfid_model == MODEL_S6500) {
        if ((err = rfid_interface_alloc(&s6500_funcs, &ifaces[iface_count++])) != 0) {
            fprintf(stderr, "%s: error: unable to allocate rfid interface: %s\n",
                    progname, strerror(err));
            exit(1);
        }
    } 

    /* Make sure we have an tag-file */
    if (argc < optind + 1) {
        fprintf(stderr, "%s: error: No tag-file named\n", progname);
        usage();
        exit(1);
    } else if (argc > optind + 1) {
        fprintf(stderr, "%s: warning: superfluous arguments ignored\n", progname);
    }

    /* Open the tag-file */
    if ((file = fopen(argv[optind], "r")) == NULL) {
        fprintf(stderr, "%s: error: unable to open tag-file %s: %s\n",
                progname, argv[optind], strerror(errno));
        exit(1);
    }

    /* Parse it */
    if (read_ids(file, &tags, &tag_count) < 0) {
        exit(1);
    }

    /* Close the file */
    if (fclose(file) < 0) {
        fprintf(stderr, "%s: warning: unable to close tag-file %s: %s\n",
                progname, argv[optind], strerror(errno));
    }

    /* Complain if we could find no valid tags */
    if (tag_count == 0) {
        fprintf(stderr, "%s: warning: no ids found in tag-file %s\n",
                progname, argv[optind]);
    }

    /* Open the serial port */
    if ((err = open_serial_port(devname, &port)) != 0) {
        fprintf(stderr, "%s: unable to open serial port %s: %s\n",
                progname, devname, strerror(err));
        exit(1);
    }

    /* Configure it now, so that we can bail early if there aren't any
     * usable configurations. */
    rfid_interface_get_settings(ifaces[iface_index], &baud, &data_bits, &parity, &stop_bits);
    if ((err = config_serial_port(port, baud, data_bits, parity, stop_bits)) != 0) {
        fprintf(stderr, "%s: unable to configure serial port: %d\n",
                progname, err);
        exit(1);
    }

#ifdef WIN32
    /* Set up the overlapped I/O stuff */
    memset(&overlapped, 0, sizeof(OVERLAPPED));
    if ((overlapped.hEvent = CreateEvent(NULL, FALSE, FALSE, NULL)) == NULL) {
        fprintf(stderr, "%s: CreateEvent failed: %d\n", progname, GetLastError());
        exit(1);
    }
#endif /* WIN32 */

    /* Connect to the Elvin server */
    dprintf(1, "%s: Connecting to the Elvin server...", progname);
    if (!elvin_async_connect(handle, connect_cb, NULL, error)) {
        if (verbosity >= 1) {
            dprintf(1, "  failed\n");
        }

        fprintf(stderr, "%s: error: ", progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    /* Run the mainloop */
    if (!elvin_mainloop_run(client, error)) {
        fprintf(stderr, "%s: error: elvin_mainloop_run failed\n%s: error: ",
                progname, progname);
        elvin_error_fprintf(stderr, error);
        exit(1);
    }

    /* Close the serial port */
    if ((err = close_serial_port(port)) != 0) {
        fprintf(stderr, "%s: warning: unable to close serial port %s: %s\n",
                progname, devname, strerror(err));
    }

    /* Free the handle */
    if (!elvin_handle_free(handle, error)) {
        fprintf(stderr, "%s: warning: elvin_handle_free failed\n%s: warning: ",
                progname, progname);
        elvin_error_fprintf(stderr, error);
    }

    /* Clean up the Elvin client */
    if (!elvin_mainloop_cleanup(client, error)) {
        fprintf(stderr, "%s: warning: elvin_mainloop_cleanup failed\n%s: warning: ",
                progname, progname);
        elvin_error_fprintf(stderr, error);
    }

    /* Free our error context */
    if (!elvin_error_free(error, NULL)) {
        fprintf(stderr, "%s: warning: elvin_error_free failed\n", progname);
    }

    exit(0);
}
