/* ids.c
 *
 * Copyright  2002, 2003, 2004 Ted Phelps
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms fo the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the `etagit' program; see the file `COPYING'.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif

#include <elvin/elvin.h>

#include "ids.h"

#define LEX_START 0
#define LEX_ID 1
#define LEX_WS 2
#define LEX_NAME 3
#define LEX_LONG_NAME 4
#define LEX_SKIP 5


#define DIGIT2INT(ch) (ch - '0')

/* Compare two tags */
static int tag_compare(const void *p1, const void *p2)
{
    const struct rfid_tag *tag1 = (const struct rfid_tag *)p1;
    const struct rfid_tag *tag2 = (const struct rfid_tag *)p2;

    return tag1 -> id < tag2 -> id ? -1 : tag1 -> id == tag2 -> id ? 0 : 1;
}

/* Reads the table mapping tag ids to names. */
int read_ids(FILE *stream, struct rfid_tag **tags_out, size_t *count_out)
{
    struct rfid_tag *tags = NULL;
    int state = LEX_START;
    char name[MAX_NAME_SIZE + 1];
    char *point = NULL;
    char *mark = NULL;
    int64_t id = 0;
    int ch;
    int line = 1;
    int count = 0;
    
    while (1)
    {
	/* Get the next character */
	if ((ch = fgetc(stream)) == EOF)
	{
	    /* Bail out on errors */
	    if (ferror(stream))
	    {
		return -1;
	    }
	}

	switch (state)
	{
	    case LEX_START:
	    {
		/* Watch for EOF */
		if (ch == EOF)
		{
		    break;
		}

		/* Remember linefeeds */
		if (ch == '\n')
		{
		    line++;
		}

		/* Ignore whitespace */
		if (isspace(ch))
		{
		    break;
		}

		/* Watch for a comment */
		if (ch == '#')
		{
		    state = LEX_SKIP;
		    break;
		}

		/* Watch for the start of a tag id */
		if (isdigit(ch))
		{
		    id = DIGIT2INT(ch);
		    state = LEX_ID;
		    break;
		}

		/* Anything else is an error */
		fprintf(stderr, "Invalid tag id on line %d\n", line);
		state = LEX_SKIP;
		break;
	    }

	    /* Skip to the end of the current line */
	    case LEX_SKIP:
	    {
		if (ch == '\n' || ch == EOF)
		{
		    line++;
		    state = LEX_START;
		}

		break;
	    }

	    /* Reading the digits of a unique id */
	    case LEX_ID:
	    {
		/* Watch for premature end of line or file */
		if (ch == '\n' || ch == EOF)
		{
		    fprintf(stderr, "Unexpected end of tag id on line %d\n", line);
		    state = LEX_START;
		    line++;
		    break;
		}

		/* Whitespace indicates the end of the id */
		if (isspace(ch))
		{
		    state = LEX_WS;
		    break;
		}

		/* More digits go into the id */
		if (isdigit(ch))
		{
		    id = 10 * id + DIGIT2INT(ch);
		    break;
		}

		/* Anything else is an error */
		fprintf(stderr, "Invalid tag id on line %d\n", line);
		state = LEX_SKIP;
		break;
	    }

	    /* Skipping whitespace between the id and the name */
	    case LEX_WS:
	    {
		/* Watch for premature end of line or file */
		if (ch == '\n' || ch == EOF)
		{
		    fprintf(stderr, "Unexpected end of tag id on line %d\n", line);
		    state = LEX_START;
		    line++;
		    break;
		}

		/* Skip more whitespace */
		if (isspace(ch))
		{
		    break;
		}

		/* Anything else is part of the name */
		point = name;
		mark = point;
		*(point++) = ch;
		state = LEX_NAME;
		break;
	    }

	    /* Reading a tag's symbolic name */
	    case LEX_NAME:
	    case LEX_LONG_NAME:
	    {
		/* Watch for end of line or file */
		if (ch == '\n' || ch == EOF)
		{
		    *mark = 0;

		    /* Make more room in the table of tags */
		    tags = (struct rfid_tag *)realloc(tags, (count + 1) * sizeof(struct rfid_tag));
		    if (tags == NULL)
		    {
			perror("realloc() failed");
			exit(1);
		    }

		    /* Add this tag */
		    tags[count].id = id;
		    strcpy(tags[count].name, name);
		    count++;

		    /* Go back to the start state for the next line */
		    state = LEX_START;
		    line++;
		    break;
		}

		/* Additional characters get added to the name */
		if (point - name < MAX_NAME_SIZE)
		{
		    *(point++) = ch;

		    /* Prepare to trim trailing whitespace */
		    if (! isspace(ch))
		    {
			mark = point;
		    }

		    break;
		}

		/* Complain about overly long names */
		if (state == LEX_NAME)
		{
		    fprintf(stderr, "Tag name too long on line %d\n", line);
		    state = LEX_LONG_NAME;
		}

		break;
	    }

	    default:
	    {
		fprintf(stderr, "Unknown state %d\n", state);
		exit(1);
	    }
	}

	/* Graceful exit? */
	if (ch == EOF && state == LEX_START)
	{
	    /* Sort the tags by id */
	    qsort(tags, count, sizeof(struct rfid_tag), tag_compare);

	    /* Return the results */
	    *count_out = count;
	    *tags_out = tags;
	    return 0;
	}
    }
}

/* Locates the name corresponding to the given id in the table */
char *find_id(struct rfid_tag *tags, size_t count, int64_t id)
{
    struct rfid_tag dummy;
    struct rfid_tag *tag;

    /* Use a dummy tag so that we can use the same compare function */
    dummy.id = id;
    if (! (tag = bsearch(&dummy, tags, count, sizeof(struct rfid_tag), tag_compare)))
    {
        return NULL;
    }

    return tag->name;
}
