/*
 * Copyright (c) 2004 Theodore A. Phelps
 *
 * eTag-It is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Twitch-o-matic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Twitch-o-matic; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_WINCONFIG_H
#include <winconfig.h>
#endif
#ifdef HAVE_WCECONFIG_H
#include <wceconfig.h>
#endif
#ifdef HAVE_WINDOWS_H
#include <windows.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#include "rfid-interface.h"

#define INITIAL_BUFFER_SIZE 64

#ifdef WIN32
#define ERRNO() GetLastError()
#else
#define ERRNO() errno
#endif


/* Allocate a new rfid-interface instance */
int rfid_interface_alloc(struct rfid_interface_funcs *funcs,
                         rfid_interface_t *interface_out)
{
    int err;
    rfid_interface_t self;

    /* Allocate memory for the interface */
    if ((self = malloc(sizeof(struct rfid_interface))) == NULL) {
        return ERRNO();
    }

    /* Record the functions table */
    self->funcs = funcs;

    /* Allocate memory for the frame construction buffer */
    if ((self->buffer = malloc(INITIAL_BUFFER_SIZE)) == NULL) {
        err = ERRNO();
        free(self);
        return err;
    }
    self->count = 0;
    self->capacity = INITIAL_BUFFER_SIZE;

    /* All is well */
    *interface_out = self;
    return 0;
}

/* Free an rfid-interface instance */
void rfid_interface_free(rfid_interface_t self)
{
    /* Free the model-specific bits */
    free(self);
}

/* Reset the rfid-interface instance */
void rfid_interface_reset(rfid_interface_t self)
{
    self->count = 0;
}

/* Determines the default serial port settings */
void rfid_interface_get_settings(rfid_interface_t self,
                                 int *baud_out,
                                 int *data_bits_out,
                                 char *parity_out,
                                 int *stop_bits_out)
{
    /* Delegate */
    self->funcs->get_settings(self, baud_out, data_bits_out, parity_out, stop_bits_out);
}


/* Build a request of the appropriate type */
int rfid_interface_make_req(rfid_interface_t self,
                            rfid_interface_type_t type,
                            char *buffer,
                            int *count_in_out)
{
    /* Hand off to the model-specific interface */
    return self->funcs->make_req(self, type, buffer, count_in_out);
}


/* Feed more bytes from the reader to the interface */
int rfid_interface_feed(rfid_interface_t self,
                        char *buffer,
                        int count,
                        int *is_complete_out,
                        rfid_response_t response_out)
{
    char *new_buffer;
    int new_capacity;

    /* Make sure there's room in the buffer */
    while (self->count + count > self->capacity) {
        new_capacity = self->capacity * 2;
        if ((new_buffer = realloc(self->buffer, new_capacity)) == NULL) {
            return ERRNO();
        }

        self->buffer = new_buffer;
        self->capacity = new_capacity;
    }

    /* Append the new bytes to the end of the buffer */
    memcpy(self->buffer + self->count, buffer, count);
    self->count += count;

    /* Hand off to the model-specific interface */
    return self->funcs->digest(self, is_complete_out, response_out);
}
