/*
 * Copyright (c) 2004 Theodore A. Phelps
 *
 * eTag-It is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Twitch-o-matic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Twitch-o-matic; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

typedef struct rfid_interface *rfid_interface_t;
typedef union rfid_response *rfid_response_t;

/* Error codes */
#define RFID_READER_ERROR -10
#define RFID_MALFORMED_FRAME -11

#if defined(HAVE_UINT64_T)
#elif defined(HAVE_UNSIGNED_LONG_LONG)
typedef unsigned long long uint64_t;
#elif defined(HAVE___INT64)
typedef __int64 uint64_t;
#else
#error Unable to determine type for uint64_t
#endif

/* The maximum length of the version string, including the null terminator */
#define RFID_VERSION_MAX 64

/* The maximum number of transponder ids that can be read */
#define RFID_MAX_IDS 50

/* The request/response types */
typedef enum
{
    /* Version request/response */
    RFID_TYPE_VERSION = 0,

    /* Inventory request/response */
    RFID_TYPE_INVENTORY,

    /* Reader reset request/response */
    RFID_TYPE_RESET,
} rfid_interface_type_t;


/* Response types */

struct rfid_interface_version_response
{
    /* The response type: RFID_TYPE_VERSION */
    rfid_interface_type_t type;

    /* The version string */
    char string[RFID_VERSION_MAX];
};

struct rfid_interface_inventory_response
{
    /* The response type: RFID_TYPE_INVENTORY */
    rfid_interface_type_t type;

    /* The number of transponders found */
    int count;

    /* The transponder UIDs */
    uint64_t ids[RFID_MAX_IDS];
};

struct rfid_interface_reset_response
{
    /* The response type: RFID_TYPE_RESET */
    rfid_interface_type_t type;
};

union rfid_response
{
    /* The result type -- must be first element of each struct */
    int type;

    /* The reader version information */
    struct rfid_interface_version_response version;
    struct rfid_interface_inventory_response inventory;
    struct rfid_interface_reset_response reset;
};


/* Function types for the interface */
typedef void (*rfid_interface_get_settings_func_t)(rfid_interface_t self,
                                                  int *baud_out,
                                                  int *data_bits_out,
                                                  char *parity_out,
                                                  int *stop_bits_out);
typedef int (*rfid_interface_make_req_func_t)(rfid_interface_t self,
                                             rfid_interface_type_t type,
                                             char *buffer,
                                             int *count_in_out);
typedef int (*rfid_interface_digest_func_t)(rfid_interface_t self,
                                            int *is_complete_out,
                                            rfid_response_t response);


/* The functions that form the interface */
struct rfid_interface_funcs
{
    /* Gets the default serial port config */
    rfid_interface_get_settings_func_t get_settings;

    /* Create a firmware version request */
    rfid_interface_make_req_func_t make_req;

    /* Consume the output of the reader */
    rfid_interface_digest_func_t digest;
};


/* The common portions of the RF-ID interface */
struct rfid_interface
{
    /* The model-specific functions */
    struct rfid_interface_funcs *funcs;

    /* Frame construction buffer */
    unsigned char *buffer;

    /* The capacity of the frame construction buffer */
    int capacity;

    /* The number of bytes in the buffer */
    int count;
};


/* The S6000 RF-ID interface functions */
extern struct rfid_interface_funcs s6000_funcs;

/* The S6500 RF-ID interface functions */
extern struct rfid_interface_funcs s6500_funcs;


/* Allocate a new rfid-interface instance */
int rfid_interface_alloc(struct rfid_interface_funcs *funcs,
                         rfid_interface_t *interface_out);

/* Free an rfid-interface instance */
void rfid_interface_free(rfid_interface_t self);

/* Reset the rfid-interface */
void rfid_interface_reset(rfid_interface_t self);

/* Determines the default serial port settings */
void rfid_interface_get_settings(rfid_interface_t self,
                                 int *baud_out,
                                 int *data_bits_out,
                                 char *parity_out,
                                 int *stop_bits_out);

/* Build a request of the appropriate type */
int rfid_interface_make_req(rfid_interface_t self,
                            rfid_interface_type_t type,
                            char *buffer,
                            int *count_in_out);

/* Feed more bytes from the reader to the interface */
int rfid_interface_feed(rfid_interface_t self,
                        char *buffer,
                        int count,
                        int *is_complete_out,
                        rfid_response_t response_out);
