/*
 * Copyright (c) 2004 Theodore A. Phelps
 *
 * eTag-It is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Twitch-o-matic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Twitch-o-matic; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_WINCONFIG_H
#include <winconfig.h>
#endif
#ifdef HAVE_WCECONFIG_H
#include <wceconfig.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include "rfid-interface.h"

#define MIN_FRAME_SIZE 6
#define MAX_FRAME_SIZE 4095
#define MAX_IDS 32
#define CRC_POLY 0x8408
#define CRC_PRESET 0xffff

#define FRAME_START 0xd5

/* The pre-defined requests.  The first byte in each array indicates
 * how many bytes follow.  The resulting request will contain 4 more
 * bytes: the length, COM-ADR (we always use 0xff) and 2 bytes of
 * CRC16 data */

/* RFID_RESET (5.15, p. 74) */
#define S6000_RESET 0x10
static unsigned char s6000_reset_payload[] = { 3, 0x04, 0x00, S6000_RESET };

/* RFID_VERSION (5.16, p. 76) */
#define S6000_VERSION 0x11
static unsigned char s6000_version_payload[] = { 3, 0x04, 0x00, S6000_VERSION };

/* RFID_INVENTORY (5.9, p. 58) */
#define S6000_INVENTORY 0xfe
static unsigned char s6000_inventory_payload[] = { 7, 0x03, 0x00, S6000_INVENTORY, 0x00, 0x00, 0x00, 0x04 };

static unsigned char *s6000_payloads[] =
{
    s6000_version_payload,
    s6000_inventory_payload,
    s6000_reset_payload
};


/* Compute the CRC of an array of bytes.  Note that this is *not* a
 * proper CCITT CRC because TI have chosen to use a common yet dodgy
 * variation on that algorithm.  For more information on how CRC-CCITT
 * should be computed, see Joe Geluso's web page at
 * http://www.joegeluso.com/software/articles/ccitt.htm */
int calc_s6000_crc(char *buffer, int length)
{
    /* Our table of CRC transforms */
    static unsigned short crc_table[256] =
    {
	0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
	0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
	0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
	0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
	0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
	0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
	0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
	0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
	0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
	0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
	0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
	0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
	0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
	0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
	0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
	0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
	0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
	0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
	0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
	0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
	0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
	0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
	0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
	0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
	0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
	0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
	0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
	0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
	0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
	0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
	0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
	0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
    };

    char *end = buffer + length;
    char *point;
    int crc;

    /* Start with the CRC preset */
    crc = CRC_PRESET;

    /* Use the table to compute the changes for each byte */
    for (point = buffer; point < end; point++) {
	crc = ((crc << 8) ^ crc_table[(crc >> 8) ^ *(unsigned char *)point]) & 0xffff;
    }

    /* For some reason we need to do this */
    return crc ^ CRC_PRESET;
}

/* Return the default serial line settings */
static void s6000_get_settings(rfid_interface_t interface,
                               int *baud_out,
                               int *data_bits_out,
                               char *parity_out,
                               int *stop_bits_out)
{
    *baud_out = 19200;
    *data_bits_out = 8;
    *parity_out = 'N';
    *stop_bits_out = 1;
}

static int s6000_make_req(rfid_interface_t interface,
                              rfid_interface_type_t type,
                              char *buffer,
                              int *count_in_out)
{
    int size;
    char *payload;
    char bytes[MAX_FRAME_SIZE];
    int crc;
    int count = *count_in_out;

    /* Look up the payload */
    payload = s6000_payloads[(int)type];
    size = *(unsigned char *)payload + 2;

    /* Header is start byte plus length */
    bytes[0] = (char)FRAME_START;
    bytes[1] = (char)((size >> 8) & 0xff);
    bytes[2] = (char)(size & 0xff);

    /* Payload */
    memcpy(bytes + 3, payload + 1, size);

    /* Compute the CRC (Appendix A, p. 89) */
    crc = calc_s6000_crc(bytes + 1, size);

    /* CRC16 */
    bytes[size + 1] = (crc >> 8) & 0xff;
    bytes[size + 2] = crc & 0xff;

    /* Copy what fits into the output buffer */
    memcpy(buffer, bytes, count < size + 3 ? count : size + 3);
    *count_in_out = size + 3;

    /* FIX THIS: use a better return code to indicate insufficient space */
    return count < size + 3 ? -1 : 0;
}

static int s6000_digest_reset(rfid_interface_t self,
                              unsigned char *buffer,
                              int size,
                              rfid_response_t response_out)
{
    response_out->reset.type = RFID_TYPE_RESET;
    return 0;
}

static int s6000_digest_version(rfid_interface_t self,
                                unsigned char *buffer,
                                int size,
                                rfid_response_t response_out)
{
    response_out->version.type = RFID_TYPE_VERSION;
    sprintf(response_out->version.string,
            "S6000 %d.%d.%d (%s)",
            buffer[6], buffer[7], buffer[8],
            buffer[9] ? "Engineering" : "Standard");
    return 0;
}

static int s6000_digest_inventory(rfid_interface_t self,
                                unsigned char *buffer,
                                int size,
                                rfid_response_t response_out)
{
    int count, i;

    response_out->inventory.type = RFID_TYPE_INVENTORY;

    /* Check the service code */
    if (buffer[3] != 0x03) {
        return RFID_READER_ERROR;
    }

    /* Check for the no transponders case */
    if (buffer[8] == 0x06) {
        response_out->inventory.count = 0;
        return 0;
    }

    /* Truncate the number of ids at the maximum */
    count = buffer[9];
    count = count < RFID_MAX_IDS ? count : RFID_MAX_IDS;

    /* Sanity check the number of ids */
    if (size < count * 4 + 10) {
        return RFID_MALFORMED_FRAME;
    }

    /* Fill in the UIDs */
    response_out->inventory.count = count;
    for (i = 0; i < count; i++) {
        response_out->inventory.ids[i] =
            (uint64_t)buffer[10 + i * 4] << 24 |
            (uint64_t)buffer[11 + i * 4] << 16 |
            (uint64_t)buffer[12 + i * 4] << 8 |
            (uint64_t)buffer[13 + i * 4];
    }

    return 0;
}

static int s6000_digest(rfid_interface_t self,
                        int *is_complete_out,
                        rfid_response_t response)
{
    int err, crc, size, i;

    /* Look for a valid frame by trying every possible offset until we
     * either find a start of frame marker and valid CRC or run out of
     * bytes */
    for (i = 0; i < self->count - MIN_FRAME_SIZE; i++) {
        /* Skip this offset if it doesn't begin with a start-of-frame */
        if (self->buffer[i] != FRAME_START) {
            continue;
        }

        /* Skip anything with an implausible size */
        size = ((self->buffer[i + 1] << 8) | self->buffer[i + 2]) + 3;
        if (size < MIN_FRAME_SIZE || size > MAX_FRAME_SIZE) {
            continue;
        }

        /* Skip sizes that are too large for what we've read thus far */
        if (self->count - i < size) {
            continue;
        }

        /* Skip frames that don't have a valid CRC */
        crc = calc_s6000_crc(self->buffer + i + 1, size - 3);
        if (self->buffer[i + size - 2] != ((crc >> 8) & 0xff) ||
            self->buffer[i + size - 1] != (crc & 0xff)) {
            continue;
        }

        /* We have a complete, verified frame -- fill in the response */
        switch (self->buffer[i + 5])
        {
            case S6000_RESET:
                err = s6000_digest_reset(self, self->buffer + i, size, response);
                break;

            case S6000_VERSION:
                err = s6000_digest_version(self, self->buffer + i, size, response);
                break;

            case S6000_INVENTORY:
                err = s6000_digest_inventory(self, self->buffer + i, size, response);
                break;

            default:
                /* FIX THIS: define some real (negative) error codes to return */
                err = -1;
                break;
        }

        /* Discard the consumed bytes */
        memmove(self->buffer, self->buffer + i + size, self->count - i - size);
        self->count -= i + size;

        /* Report errors */
        if (err) {
            return err;
        }

        *is_complete_out = 1;
        return 0;
    }

    /* If the frame is now larger than the maximum frame size then
     * discard the bytes at the beginning so that the buffer doesn't
     * grow without bound. */
    if (self->count > MAX_FRAME_SIZE) {
        memmove(self->buffer,
                self->buffer + self->count - MAX_FRAME_SIZE,
                MAX_FRAME_SIZE);
        self->count -= MAX_FRAME_SIZE;
    }

    /* Not enough bytes for a whole frame */
    *is_complete_out = 0;
    return 0;
}


/* The S6000 RF-ID tag reader interface functions */
struct rfid_interface_funcs s6000_funcs =
{
    s6000_get_settings,
    s6000_make_req,
    s6000_digest
};

