/*
 * Copyright (c) 2004 Theodore A. Phelps
 *
 * eTag-It is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Twitch-o-matic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Twitch-o-matic; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_WINCONFIG_H
#include <winconfig.h>
#endif
#ifdef HAVE_WCECONFIG_H
#include <wceconfig.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include "rfid-interface.h"

#define MIN_FRAME_SIZE 5
#define MAX_FRAME_SIZE 258
#define CRC_POLY 0x8408
#define CRC_PRESET 0xffff

/* The pre-defined requests.  The first byte in each array indicates
 * how many bytes follow.  The resulting request will contain 4 more
 * bytes: the length, COM-ADR (we always use 0xff) and 2 bytes of
 * CRC16 data  */

/* RFID_RESET (5.6, p. 57) */
#define S6500_RESET 0x63
static unsigned char s6500_reset_payload[] = { 1, S6500_RESET };

/* RFID_VERSION (5.7, p. 57) */
#define S6500_VERSION 0x65
static unsigned char s6500_version_payload[] = { 1, S6500_VERSION };

/* RFID_INVENTORY (6.2.1, p. 68) */
#define S6500_INVENTORY 0xb0
static unsigned char s6500_inventory_payload[] = { 3, S6500_INVENTORY, 0x01, 0x00 };

static unsigned char *s6500_payloads[] =
{
    s6500_version_payload,
    s6500_inventory_payload,
    s6500_reset_payload
};


/* Computes a CRC16 using the S6500 polynomial and initial value */
static int calc_s6500_crc(char *buffer, int count)
{
    int crc, i, j;

    crc = CRC_PRESET;
    for (i = 0; i < count; i++) {
        crc ^= *(unsigned char *)(buffer + i);
        for (j = 0; j < 8; j++) {
            if ((crc & 1) == 0) {
                crc >>= 1;
            } else {
                crc = (crc >> 1) ^ CRC_POLY;
            }
        }
    }

    return crc;
}

/* Return the default serial line settings */
static void s6500_get_settings(rfid_interface_t interface,
                               int *baud_out,
                               int *data_bits_out,
                               char *parity_out,
                               int *stop_bits_out)
{
    *baud_out = 38400;
    *data_bits_out = 8;
    *parity_out = 'E';
    *stop_bits_out = 1;
}

static int s6500_make_req(rfid_interface_t interface,
                              rfid_interface_type_t type,
                              char *buffer,
                              int *count_in_out)
{
    int size;
    char *payload;
    char bytes[MAX_FRAME_SIZE];
    int crc;
    int count = *count_in_out;

    /* Look up the payload */
    payload = s6500_payloads[(int)type];
    size = *(unsigned char *)payload;

    /* Create the frame: length of frame */
    bytes[0] = size + 4;

    /* COM-ADR */
    bytes[1] = (char)0xff;

    /* Payload */
    memcpy(bytes + 2, payload + 1, size);

    /* Compute the CRC (2.5, p. 18) */
    crc = calc_s6500_crc(bytes, size + 2);

    /* CRC16 */
    bytes[size + 2] = crc & 0xff;
    bytes[size + 3] = (crc >> 8) & 0xff;

    /* Copy what fits into the output buffer */
    memcpy(buffer, bytes, count < size + 4 ? count : size + 4);
    *count_in_out = size + 4;

    /* FIX THIS: use a better return code to indicate insufficient space */
    return count < size + 4 ? -1 : 0;
}

static int s6500_digest_reset(rfid_interface_t self,
                              unsigned char *buffer, int size,
                              rfid_response_t response_out)
{
    response_out->reset.type = RFID_TYPE_RESET;
    return 0;
}

static int s6500_digest_version(rfid_interface_t self,
                                unsigned char *buffer, int size,
                                rfid_response_t response_out)
{
    response_out->version.type = RFID_TYPE_VERSION;
    sprintf(response_out->version.string,
            "S6500 %d.%d (%d)",
            buffer[4] & 0xf,
            buffer[5],
            buffer[4] >> 4);
    return 0;
}

static int s6500_digest_inventory(rfid_interface_t self,
                                  unsigned char *buffer, int size,
                                  rfid_response_t response_out)
{
    int count, i;

    response_out->inventory.type = RFID_TYPE_INVENTORY;

    /* Check for the no transponders case */
    if (buffer[3] == 0x01) {
        response_out->inventory.count = 0;
        return 0;
    }

    /* Watch for random status errors */
    if (buffer[3] != 0x00 && buffer[3] != 0x94) {
        return RFID_READER_ERROR;
    }

    /* Truncate the number of ids at the maximum */
    count = buffer[4];
    count = count < RFID_MAX_IDS ? count : RFID_MAX_IDS;

    /* Sanity check the number of ids */
    if (size < count * 10 + 7) {
        return RFID_MALFORMED_FRAME;
    }

    /* Fill in the UIDs */
    response_out->inventory.count = count;
    for (i = 0; i < count; i++) {
        response_out->inventory.ids[i] =
            (uint64_t)buffer[5 + i * 10 + 2] << 56 |
            (uint64_t)buffer[5 + i * 10 + 3] << 48 |
            (uint64_t)buffer[5 + i * 10 + 4] << 40 |
            (uint64_t)buffer[5 + i * 10 + 5] << 32 |
            (uint64_t)buffer[5 + i * 10 + 6] << 24 |
            (uint64_t)buffer[5 + i * 10 + 7] << 16 |
            (uint64_t)buffer[5 + i * 10 + 8] << 8 |
            (uint64_t)buffer[5 + i * 10 + 9];
    }

    return 0;
}

static int s6500_digest(rfid_interface_t self,
                        int *is_complete_out,
                        rfid_response_t response_out)
{
    int err, crc, size, i;

    /* Look for a valid frame by trying every possible offset until we
     * either find a valid CRC or run out of bytes */
    for (i = 0; i < self->count - MIN_FRAME_SIZE; i++) {
        /* Skip sizes that are implausibly small */
        if ((size = self->buffer[i]) < MIN_FRAME_SIZE) {
            continue;
        }

        /* Skip sizes that are too large for what we've read thus far */
        if (self->count - i < size) {
            continue;
        }

        /* Skip sizes that don't have a valid CRC.  Note that we may
         * end up computing a few extra CRCs, but in practise there
         * just aren't that many non-size bytes between MIN_FRAME_SIZE
         * and self->count */
        crc = calc_s6500_crc(self->buffer + i, size - 2);
        if (self->buffer[i + size - 2] != (crc & 0xff) ||
            self->buffer[i + size - 1] != ((crc >> 8) & 0xff)) {
            continue;
        }


        /* We have a valid frame.  Parse it and write the info into
         * the response */
        switch (self->buffer[i + 2])
        {
            case S6500_RESET:
                err = s6500_digest_reset(self,
                                         self->buffer + i, size,
                                         response_out);
                break;

            case S6500_VERSION:
                err = s6500_digest_version(self,
                                           self->buffer + i, size,
                                           response_out);
                break;

            case S6500_INVENTORY:
                err = s6500_digest_inventory(self,
                                             self->buffer + i, size,
                                             response_out);
                break;
                
            default:
                /* FIX THIS: define some real (negative) error codes
                 * to return */
                err = -1;
                break;
        }

        /* Discard the consumed bytes */
        memmove(self->buffer, self->buffer + i + size, self->count - i - size);
        self->count -= i + size;

        /* Report errors */
        if (err) {
            return err;
        }

        *is_complete_out = 1;
        return 0;
    }

    /* If the frame is now larger than the maximum frame size then
     * discard the bytes at the beginning so that the buffer doesn't
     * grow without bound. */
    if (self->count > MAX_FRAME_SIZE) {
        memmove(self->buffer,
                self->buffer + self->count - MAX_FRAME_SIZE,
                MAX_FRAME_SIZE);
        self->count -= MAX_FRAME_SIZE;
    }

    /* No valid frame yet */
    *is_complete_out = 0;
    return 0;
}


/* The S6500 RF-ID tag reader interface functions */
struct rfid_interface_funcs s6500_funcs =
{
    s6500_get_settings,
    s6500_make_req,
    s6500_digest
};

