/* Cyclic Redundancy Check algorithm, complete with traditional error.
 * Copyright  2002, 2003, 2004 Ted Phelps
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms fo the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the `etagit' program; see the file `COPYING'.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_TIMEB_H
#include <sys/timeb.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include <elvin/elvin.h>
#include "hexdump.h"
#include "ids.h"

/* The types of readers supported */
#define S6000 6000
#define S6500 6500

/* The size of the input buffer */
#define BUFFER_SIZE 4096

/* The minimum number of bytes to read before we have the packet length */
#define FRAME_SIZE 3

/* The maximum number of IDs we're willing to track */
#define MAX_IDS 256

/* The maximum number of antennas attached to the reader */
#define MAX_ANTENNAS 2
static int antenna_count = 0;
static char *sources[MAX_ANTENNAS];

#ifdef WIN32
#define DEFAULT_DEVICE "COM1"
#else /* WIN32 */
#define DEFAULT_DEVICE "/dev/ttyS0"
#endif /* WIN32 */

/* Bytes read from the reader */
static char read_buffer[BUFFER_SIZE];
static char *read_end = read_buffer + BUFFER_SIZE;
static char *read_point = read_buffer;

/* Record of the IDs near the antenna */
static uint64_t current_ids[MAX_ANTENNAS][MAX_IDS];
static int current_id_count[MAX_ANTENNAS];

/* The current antenna */
static int antenna = 0;

/* Elvin information*/
static elvin_client_t client = NULL;
static elvin_error_t error = NULL;
static elvin_handle_t handle = NULL;

/* How long to wait between requests to the reader */
#define DEFAULT_POLL_PERIOD 250
static int poll_period = DEFAULT_POLL_PERIOD;

/* How much spewage to generate */
static int verbosity = 0;

/* Add in the WOOD field for the ambient wood */
static int wood_hack = 0;

/* Have we found a reader? */
static int found_reader = 0;

/* The table mapping tag ids to names */
static struct rfid_tag *tags;
size_t tag_count;

#if defined(KEYBOARD_CONTROL)
static HANDLE console;
#endif

#ifdef WIN32
extern char *optarg;
static HANDLE port = NULL;
#else /* WIN32 */
static int fd = -1;
static elvin_timeout_t request_timeout = NULL;
#endif /* WIN32 */

struct reader_info
{
    /* How long to wait for a reply from the reader */
    int request_timeout;

    /* Get the reader version */
    unsigned char req_reader_version[20];
    int req_reader_version_size;

    /* Read the current tag-ids */
    unsigned char req_read_ids[20];
    int req_read_ids_size;

    /* Switch to antenna 0 */
    unsigned char req_set_output0[20];
    int req_set_output0_size;

    /* Switch to antenna 1 */
    unsigned char req_set_output1[20];
    int req_set_output1_size;
};

/* The current reader's info */
struct reader_info *reader;

/* The reader info for an S6000 */
struct reader_info s6000_info =
{
    /* Request timeout */
    1500,

    /* Get version */
    { 0xd5, 0x00, 0x05, 0x04, 0x00, 0x11, 0x8c, 0x66 }, 8,

    /* Read block SID */
    { 0xd5, 0x00, 0x09, 0x03, 0x00, 0xfe, 0x00, 0x00, 0x01, 0x01, 0xb5, 0xe8 }, 12,

    /* Antenna 0 */
    { 0 }, 0,

    /* Antenna 1 */
    { 0 }, 0
};

/* The reader info for an S6500 */
struct reader_info s6500_info = 
{
    /* Request timeout */
    10000,

    /* Get Software Version */
    { 0x05, 0xff, 0x65, 0xe5, 0xcb }, 5,

    /* Inventory */
    { 0x07, 0xff, 0xb0, 0x01, 0x00, 0x1c, 0x56 }, 7,

#if defined(USE_RELAY)
    /* Set Output (relay=inactive) */
    { 0x0d, 0xff, 0x71, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x28, 0x68, 0xa2 }, 13,

    /* Set Output (relay=1) */
    { 0x0d, 0xff, 0x71, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x28, 0x7b, 0x0c }, 13
#else /* ! USE_RELAY */
    /* Set Output (relay=inactive) */
    { 0x0d, 0xff, 0x71, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x28, 0x36, 0x44 }, 13,

    /* Set Output (relay=1) */
    { 0x0d, 0xff, 0x71, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x28, 0x54, 0x7f }, 13
#endif
};

static int send_request(char *buffer, size_t size)
{
    char *point;
    char *end;
#ifdef WIN32
    DWORD length;
#else
    ssize_t length;
#endif

    /* Write a request to the serial port */
    point = buffer;
    end = buffer + size;

    if (2 < verbosity)
    {
	if (3 < verbosity)
	{
	    printf("request block:\n");
	    fhexdump(point, end - point, stdout);
	}

	printf("writing request block..");
	fflush(stdout);
    }

    while (point < end)
    {
	if (2 < verbosity)
	{
	    printf("."); fflush(stdout);
	}

#ifdef WIN32
	if (! WriteFile(
		port, point, end - point,
		&length, NULL))
	{
	    fprintf(stderr, "WriteFile() failed: %s\n",
		    strerror(GetLastError()));
	    return -1;
	}
#else
	if ((length = write(fd, point, end - point)) < 0) {
	    perror("write() failed");
	    return -1;
	}
#endif

	point += length;
    }

    if (2 < verbosity)
    {
	printf("  done\n");
    }

    return 0;
}

#ifndef WIN32
/* Forward declaration */
static int serial_port_timeout_cb(
    elvin_timeout_t timeout,
    void *rock,
    elvin_error_t error);

/* The timeout callback for a Serial-port request */
static int serial_port_request_timeout_cb(
    elvin_timeout_t timeout,
    void *rock,
    elvin_error_t error)
{
    /* Log failure */
    if (1 < verbosity)
    {
	printf("request timed out\n");
    }

    /* Reset the input buffer */
    read_point = read_buffer;

    /* Start the cycle again */
    if (! serial_port_timeout_cb(NULL, rock, error))
    {
	return 0;
    }

    return 1;
}

/* The timeout callback for polling the serial port */
static int serial_port_timeout_cb(
    elvin_timeout_t timeout,
    void *rock,
    elvin_error_t error)
{
    /* Set a timeout for the request */
    if (! (request_timeout = elvin_mainloop_add_timeout(
	       client,
	       reader->request_timeout,
	       serial_port_request_timeout_cb,
	       NULL,
	       error)))
    {
	fprintf(stderr, "elvin_mainloop_add_timeout() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Write a request to the serial port */
    send_request(reader->req_read_ids, reader->req_read_ids_size);
    return 1;
}
#endif

#if defined(WIN32)
/* Returns the number of milliseconds since the epoch */
int64_t get_timestamp()
{
    struct timeb now;

    ftime(&now);

    return (int64_t)now.time * 1000 + (int64_t)now.millitm;
}
#else /* WIN32 */
int64_t get_timestamp()
{
    struct timeval now;

    if (gettimeofday(&now, NULL) < 0)
    {
	perror("gettimeofday() failed");
	exit(1);
    }

    return (int64_t)now.tv_sec * 1000 + (int64_t)now.tv_usec;
}
#endif /* WIN32 */

/* Send a tag added notification */
static void send_notification(int antenna, uint64_t id, int was_added)
{
    elvin_notification_t notification;
    char *name;

    /* Create a notification */
    if (! (notification = elvin_notification_alloc(client, error))) {
	fprintf(stderr, "elvin_notification_alloc() failed");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Look up the tag's name */
    if ((name = find_id(tags, tag_count, id)) == NULL)
    {
	name = "unknown tag";
    }

    if (0 < verbosity)
    {
	printf("Source=\"%s\", Tag-Id=%" ELVIN_UINT64_PRINT
	       ", Name=\"%s\", Action=\"%s\"\n",
	       sources[antenna], id, name,
	       was_added ? "added" : "removed");
    }

    /* Add some fields to it */
    if (wood_hack)
    {
	if (! elvin_notification_add_int64(notification, "WOOD", get_timestamp(), error))
	{
	    fprintf(stderr, "elvin_notification_add_int64() failed\n");
	    elvin_error_fprintf(stderr, error);
	    exit(1);
	}
    }

    if (! elvin_notification_add_string(notification, "Source", sources[antenna], error) ||
	! elvin_notification_add_int64(notification, "Tag-Id", id, error) ||
	! elvin_notification_add_string(notification, "Name", name, error) ||
	! elvin_notification_add_string(notification, "Action", was_added ? "added" : "removed", error)) {
	fprintf(stderr, "elvin_notification_add_xxx() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Send it */
    if (! elvin_sync_notify(handle, notification, 1, NULL, error)) {
	fprintf(stderr, "elvin_sync_notify() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }
}

/* Send notifications in response to changes in the inventory */
static void update_inventory(uint64_t *ids, int count)
{
    int i, j;

    /* Dump the RFIDs if the verbosity level is high enough */
    if (1 < verbosity)
    {
	printf("last: [%d]", current_id_count[antenna]);
	for (j = 0; j < current_id_count[antenna]; j++)
	{
	    printf(" %" ELVIN_UINT64_PRINT , current_ids[antenna][j]);
	}

	printf("  this: [%d]", count);
	for (i = 0; i < count; i++)
	{
	    printf(" %" ELVIN_UINT64_PRINT, ids[i]);
	}
	printf("\n");
    }

    /* Figure out which ids have been added since last time */
    for (i = 0; i < count; i++)
    {
	int found = 0;

	for (j = 0; j < current_id_count[antenna]; j++)
	{
	    if (ids[i] == current_ids[antenna][j])
	    {
		current_ids[antenna][j] = 0;
		found = 1;
		break;
	    }
	}

	if (! found)
	{
	    send_notification(antenna, ids[i], 1);
	}
    }

    /* Anything left has been removed */
    for (j = 0; j < current_id_count[antenna]; j++)
    {
	if (current_ids[antenna][j] != 0)
	{
	    send_notification(antenna, current_ids[antenna][j], 0);
	}
    }

    memcpy(current_ids[antenna], ids, count * sizeof(uint64_t));
    current_id_count[antenna] = count;
}

/* Interpret a frame read from the serial port */
static void interpret_frame(unsigned char *buffer, ssize_t length)
{
    uint64_t ids[MAX_IDS];

    if (2 < verbosity)
    {
	printf("read %d byte frame\n", length);
    }

    switch (buffer[2])
    {
	/* Did we get an S6000 version reply? */
	case 0x11:
	{
	    if (0 < verbosity)
	    {
		printf("Firmware version %d.%d.%d Type: %s\n",
		       buffer[3], buffer[4], buffer[5],
		       buffer[6] ? "Engineering" : "Standard");
	    }

	    found_reader = 1;
	    return;
	}

	/* Did we get an S6500 version reply? */
	case 0x65:
	{
	    if (0 < verbosity)
	    {
		printf("Reader Version:\n");
		printf("  COM-ADR: %d\n", buffer[1]);
		printf("  Status: %d\n", buffer[3]);
		printf("  Software Revision: %02x%02x\n", buffer[4], buffer[5]);
		printf("  etc\n");
	    }

	    found_reader = 1;
	    return;
	}

	/* Did we get an S6500 Set Output reply? */
	case 0x71:
	{
	    /* Check the status code */
	    if (buffer[3] != 0)
	    {
		fprintf(stderr, "Set Output: status=0x%02x\n", buffer[3]);
	    }

	    return;
	}

	/* Did we get an S6500 inventory reply? */
	case 0xb0:
	{
	    int count, i;

	    /* Check the status code */
	    switch (buffer[3])
	    {
		/* OK */
		case 0x00:
		{
		    /* Look up the number of transponders */
		    count = buffer[4];
		    break;
		}

		/* No transponder */
		case 0x01:
		{
		    count = 0;
		    break;
		}

		/* Trouble */
		default:
		{
		    fprintf(stderr, "status problem: 0x%02x\n", buffer[3]);
		    return;
		}
	    }

	    /* Pull them out */
	    for (i = 0; i < count; i++)
	    {
		unsigned char *point = buffer + 7 + i * 10;
		ids[i] = 0;

		for (point = buffer + 7 + i * 10; point < buffer + 15 + i * 10; point++)
		{
		    ids[i] = (ids[i] << 8) | *point;
		}
	    }

	    /* Determine what has changed */
	    update_inventory(ids, count);
	    return;
	}

	/* Make sure we've got an SID read packet */
	case 0xfe:
	{
	    int i, count = 0;

	    /* Pull the RFIDs out of the packet */
	    for (i = 7; i + 7 < length; i += 10)
	    {
		ids[count++] = buffer[i] << 24 | buffer[i + 1] << 16 | buffer[i + 2] << 8 | buffer[i + 3];
	    }

	    update_inventory(ids, count);
	    return;
	}

	default:
	{
	    fprintf(stderr, "unexpected frame type: %02x\n", buffer[2]);
	    found_reader = 0;
	}
    }
}

/* Split the input buffer into frames.  Returns the number of frames found. */
static int interpret_stream()
{
    unsigned char *ptr;
    int length;
    int count = 0;

    /* Digest all of the complete frames we've read thus far */
    ptr = (unsigned char *)read_buffer;
    while (! ((unsigned char *)read_point - ptr < FRAME_SIZE))
    {
	/* Check for the S6000 start code */
	if (ptr[0] == 0xd5)
	{
	    /* The next two bytes are the length */
	    length = (ptr[1] << 8) | ptr[2];

	    /* Bail out now if we didn't get the whole frame */
	    if (read_point - (char *)ptr < length + 3)
	    {
		return count;
	    }

	    /* Interpret the contents of the frame */
	    interpret_frame(ptr + 3, length);

	    /* Move on to the next frame */
	    ptr += length + 3;
	}
	else /* S6500 */
	{
	    /* Get the length byte */
	    length = *(unsigned char *)ptr;

	    /* Bail out now if we didn't get the whole frame */
	    if (read_point - (char *)ptr < length)
	    {
		return count;
	    }

	    /* Interpret the contents of the frame */
	    interpret_frame(ptr, length);

	    /* Move on to the next frame */
	    ptr += length;
	}

	count++;
    }

    /* Move any partially read frame to the beginning of the buffer */
    memmove(read_buffer, (char *)ptr, read_point - (char *)ptr);
    read_point -= ((char *)ptr - read_buffer);
    return count;
}

/* Returns non-zero if a complete reply was read */
static int read_reply()
{
#ifdef WIN32
    DWORD length;
#else
    ssize_t length;
#endif

    /* The serial port has data for us */
    if (2 < verbosity)
    {
	printf("reading reply...");
	fflush(stdout);
    }

#ifdef WIN32
    if (! ReadFile(
	    port,
	    read_point, read_end - read_point,
	    &length, NULL))
    {
	COMSTAT comstat;
	DWORD err = GetLastError();

	/* Why have we started getting these errors? */
	if (err == ERROR_OPERATION_ABORTED)
	{
	    if (2 < verbosity)
	    {
		printf("  failed\n");
	    }

	    fprintf(stderr, "ReadFile() aborted.  Restarting...\n");

	    /* Clear the error */
	    if (! ClearCommError(port, &err, &comstat))
	    {
		fprintf(stderr, "ClearCommError() failed: %s\n",
			strerror(GetLastError()));
		return -1;
	    }

	    printf("err=0x%x\n", err);
	    read_point = read_buffer;
	    return 1;
	}

	fprintf(stderr, "ReadFile() failed: %d: %s\n", err, strerror(err));
	return -1;
    }
#else /* ! WIN32 */
    if ((length = read(fd, read_point, read_end - read_point)) < 0)
    {
	perror("read() failed");
	return -1;
    }
#endif

    if (2 < verbosity)
    {
	printf("  done: read %d bytes (buffer=%p size=%d)\n",
	       length, read_point, read_end - read_point);
    }

    if (3 < verbosity)
    {
	fhexdump(read_point, length, stdout);
    }

    read_point += length;

    /* Watch for empty reads */
    if (length == 0)
    {
#ifdef WIN32
	if (3 < verbosity)
	{
	    printf("zero-length read\n");
	}

	/* Reset the input buffer */
	read_point = read_buffer;
	return 1;
#else
	/* Terminal closed?! */
	fprintf(stderr, "unexpected end of input\n");
	return -1;
#endif
    }

    /* Digest what we've read and return if we've received a
     * complete frame */
    return interpret_stream();
}

#ifdef WIN32
#if defined(KEYBOARD_CONTROL)
static void wait_keyboard(void)
{
    char buffer[BUFFER_SIZE];
    int count;

    printf("Press Enter:");
    fflush(stdout);

    /* Read the CR */
    if (! ReadFile(console, buffer, BUFFER_SIZE, &count, NULL))
    {
	perror("ReadFile() failed");
	exit(1);
    }
}
#endif /* KEYBOARD_CONTROL */

static DWORD WINAPI do_tagit(LPVOID rock)
{
    int result;

    /* No reader found yet */
    found_reader = 0;

    /* Never give up */
    while (1)
    {
        /* Look for the RF tag-reader */
        if (0 < verbosity)
        {
            printf("Looking for RF tag-reader\n");
        }

        while (! found_reader)
        {
            /* Request the tag reader's version information */
            if (send_request(reader->req_reader_version, reader->req_reader_version_size) < 0)
            {
                continue;
            }

            /* Wait for a reply */
            while ((result = read_reply()) == 0);
            if (result < 0)
            {
                found_reader = 0;
                continue;
            }
        }

        /* Keep on truckin */
        while (found_reader)
        {
            if (reader -> req_set_output0_size)
            {
                /* Toggle the flag */
                antenna = (antenna + 1) % antenna_count;

                /* Toggle the relay */
                if (antenna)
                {
                    if (1 < verbosity)
                    {
                        printf("Switching to antenna 0\n");
                    }

                    if (send_request(reader->req_set_output0, reader->req_set_output0_size) < 0)
                    {
                        found_reader = 0;
                        continue;
                    }
                }
                else
                {
                    if (1 < verbosity)
                    {
                        printf("Switching to antenna 1\n");
                    }

                    if (send_request(reader->req_set_output1, reader->req_set_output1_size) < 0)
                    {
                        found_reader = 0;
                        continue;
                    }
                }

                /* Wait for a reply */
                while ((result = read_reply()) == 0);
                if (result < 0)
                {
                    found_reader = 0;
                    continue;
                }

#if defined(KEYBOARD_CONTROL)
                /* Read a linefeed from the console */
                wait_keyboard();
#else
                /* Sleep for a bit */
                Sleep(poll_period);
#endif /* KEYBOARD_CONTROL */
            }

            /* Send an inventory request to the reader */
            if (send_request(reader->req_read_ids, reader->req_read_ids_size) < 0)
            {
                found_reader = 0;
                continue;
            }

            /* Wait for a reply */
            while ((result = read_reply()) == 0);
            if (result < 0)
            {
                found_reader = 0;
                continue;
            }

#if defined(KEYBOARD_CONTROL)
            wait_keyboard();
#else /* ! KEYBOARD_CONTROL */
            /* Sleep for a bit */
            Sleep(poll_period);
#endif /* KEYBOARD_CONTROL */
        }
    }
}
#else /* !WIN32 */
/* The read I/O handler for the serial port */
static int serial_port_read_cb(elvin_io_handler_t handler,
			       elvin_socket_t fd,
                               uint32_t active_mask,
                               int *would_block_out,
			       void *rock,
			       elvin_error_t error)
{
    /* Read what's available and break it into frames */
    if (! read_reply()) {
	/* Bail if no complete frames read */
	return 1;
    }

    /* Cancel the reply timeout */
    if (request_timeout)
    {
	if (! elvin_mainloop_remove_timeout(client, request_timeout, error))
	{
	    fprintf(stderr, "elvin_mainloop_remove_timeout() failed\n");
	    elvin_error_fprintf(stderr, error);
	    exit(1);
	}

	request_timeout = NULL;
    }

    /* Wait a bit before reading again */
    if (! elvin_mainloop_add_timeout(client,
				     poll_period,
				     serial_port_timeout_cb,
				     NULL,
				     error)) {
	fprintf(stderr, "elvin_mainloop_add_timeout() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    return 1;
}
#endif

/* This is called when our connection attempt has completed */
static int connect_cb(
    elvin_handle_t handle,
    int result,
    void *rock,
    elvin_error_t error)
{
#ifdef WIN32
    DWORD thread_id;

    fprintf(stderr, "  done\n");

    /* Spawn a separate thread to play with the tag reader */
    if (CreateThread(NULL, 0, do_tagit, NULL, 0, &thread_id) == NULL)
    {
	fprintf(stderr, "Unable to spawn tag-reader thread\n");
	exit(1);
    }
#else /* ! WIN32 */
    fprintf(stderr, "  done\n");

    /* Add an I/O handler to cope with input from the tag reader */
    if (! elvin_mainloop_add_io_handler(
	    client, fd, ELVIN_READ_MASK,
	    serial_port_read_cb, handle,
	    error))
    {
	fprintf(stderr, "elvin_mainloop_add_io_handler() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Arrange to poll the tag reader in a little while */
    if (! elvin_mainloop_add_timeout(
	    client, poll_period,
	    serial_port_timeout_cb, NULL,
	    error)) {
	fprintf(stderr, "  failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }
#endif /* WIN32 */

    return 1;
}

/* Print a helpful usage message */
static void usage(int argc, char *argv[])
{
    fprintf(stderr,
	    "usage: %s [OPTION]... tag-file\n"
	    "  -e elvin-url\n"
	    "  -S scope\n"
	    "  -d device\n"
	    "  -a antenna\n"
	    "  -p milliseconds\n"
	    "  -s 6000|6500\n"
	    "  -w\n"
	    "  -v\n"
	    "  -h\n"
	    "  -V\n",
	    argv[0]);
}

#ifdef WIN32
/* Initialize the serial port */
static HANDLE init_serial_port(LPCSTR device, int type)
{
    HANDLE port;
    DCB dcb;
    COMMTIMEOUTS timeouts;
    fprintf(stderr, "opening %s...", device);
    fflush(stderr);

    /* Open the COM port */
    if ((port = CreateFile(
	     device,
	     GENERIC_READ | GENERIC_WRITE, 0,
	     NULL, OPEN_EXISTING,
	     0, NULL)) == INVALID_HANDLE_VALUE)
    {
	fprintf(stderr, "  failed: %s\n", strerror(GetLastError()));
	exit(1);
    }
    fprintf(stderr, "  done\n");

    /* Get the current settings */
    fprintf(stderr, "Configuring the port..."); fflush(stderr);
    if (! GetCommState(port, &dcb))
    {
	fprintf(stderr, "  failed: %s\n", strerror(GetLastError()));
	exit(1);
    }

    /* Set the baud rate, character size, parity and stop bits */
    if (type == S6000)
    {
        dcb.DCBlength = sizeof(dcb);
        dcb.BaudRate = CBR_19200;
        dcb.fBinary = TRUE;
        dcb.fParity = TRUE;
        dcb.fNull = FALSE;
        dcb.fAbortOnError = TRUE;
        dcb.ByteSize = 8;
        dcb.Parity = NOPARITY;
        dcb.StopBits = ONESTOPBIT;
    }
    else if (type == S6500)
    {
        dcb.DCBlength = sizeof(dcb);
        dcb.BaudRate = CBR_38400;
        dcb.fBinary = TRUE;
        dcb.fParity = TRUE;
        dcb.fNull = FALSE;
        dcb.fAbortOnError = TRUE;
        dcb.ByteSize = 8;
        dcb.Parity = EVENPARITY;
        dcb.StopBits = ONESTOPBIT;
    }
    else
    {
        fprintf(stderr, "Unsupported reader type %d\n", type);
        exit(1);
    }

    /* Write them back */
    if (! SetCommState(port, &dcb))
    {
	fprintf(stderr, "  failed: %s\n", strerror(GetLastError()));
	exit(1);
    }

    /* Set the COMM timeouts */
    timeouts.ReadIntervalTimeout = MAXDWORD;
    timeouts.ReadTotalTimeoutMultiplier = MAXDWORD;
    timeouts.ReadTotalTimeoutConstant = reader->request_timeout;
    timeouts.WriteTotalTimeoutMultiplier = 0;
    timeouts.WriteTotalTimeoutConstant = 0;

    if (! SetCommTimeouts(port, &timeouts))
    {
	fprintf(stderr, "  failed: %s\n", strerror(GetLastError()));
	exit(1);
    }

    /* Flush the buffer */
    if (! PurgeComm(
	    port,
	    PURGE_TXABORT | PURGE_RXABORT |
	    PURGE_TXCLEAR | PURGE_RXCLEAR))
    {
	fprintf(stderr, "  failed: %s\n", strerror(GetLastError()));
	exit(1);
    }
    fprintf(stderr, "  done\n");
    return port;
}
#else /* !WIN32 */
static int init_serial_port(char *device, int type)
{
    struct termios termios;
    int fd;

    /* Set up our preferred settings */
    memset(&termios, 0, sizeof(struct termios));
    if (type == S6000)
    {
        termios.c_cflag = B19200 | CS8 | CLOCAL | CREAD;
    }
    else if (type == S6500)
    {
        termios.c_cflag = B38400 | CS8 | CLOCAL | CREAD | PARENB;
    }

    /* Open the serial device */
    fprintf(stderr, "opening %s...", device); fflush(stderr);
    if ((fd = open(device, O_RDWR | O_NOCTTY)) < 0) {
	fprintf(stderr, "  failed: %s\n", strerror(errno));
	exit(1);
    }
    fprintf(stderr, "  done\n");

    /* Clear the serial line */
    if (tcflush(fd, TCIFLUSH) < 0) {
	perror("tcflush() failed");
	exit(1);
    }

    /* Activate our new serial line settings */
    if (tcsetattr(fd, TCSANOW, &termios) < 0) {
	perror("tcsetattr() failed");
	exit(1);
    }

    return fd;
}
#endif /* WIN32 */


/* Parse the command line, set up the connection to the tag reader and
 * the elvin server and go */
int main(int argc, char *argv[])
{
    char *device = DEFAULT_DEVICE;
    FILE *stream;
    int choice;
    int type;

    /* Default to an S6000 */
    type = S6000;
    reader = &s6000_info;

    /* Get an elvin error context */
    if (! (error = elvin_error_alloc(NULL, NULL)))
    {
	fprintf(stderr, "elvin_error_alloc() failed");
	exit(1);
    }

#ifdef WIN32
    if (! (client = elvin_win32_init_default(error)))
    {
	fprintf(stderr, "elvin_win32_init_default() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }
#else /* WIN32 */
    /* Initialize the synchronous API */
    if (! (client = elvin_default_init_default(error)))
    {
	fprintf(stderr, "elvin_default_init_default() failed");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }
#endif /* WIN32 */

    /* Allocate a connection handle */
    if (! (handle = elvin_handle_alloc(client, error)))
    {
	fprintf(stderr, "elvin_handle_alloc() failed");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Read command-line options */
    while (1)
    {
	/* Watch for the end of the options */
	if ((choice = getopt(argc, argv, "a:e:S:d:hp:s:Vvw")) < 0) {
	    break;
	}

	switch (choice)
	{
	    case 'a':
	    {
		if (! (antenna_count < MAX_ANTENNAS))
		{
		    fprintf(stderr, PACKAGE ": too many antennas\n");
		    usage(argc, argv);
		    exit(1);
		}

		sources[antenna_count++] = optarg;
		break;
	    }

	    case 'e':
	    {
		if (! elvin_handle_append_url(handle, optarg, error)) {
		    fprintf(stderr, PACKAGE ": invalid URL: \"%s\"\n", optarg);
		    elvin_error_fprintf(stderr, error);
		    exit(1);
		}

		break;
	    }

	    case 'S':
	    {
		if (! elvin_handle_set_discovery_scope(handle, optarg, error)) {
		    fprintf(stderr, "elvin_handle_set_discovery_scope() failed\n");
		    elvin_error_fprintf(stderr, error);
		    exit(1);
		}

		break;
	    }

	    case 'd':
	    {
		device = optarg;
		break;
	    }

	    case 'p':
	    {
		poll_period = atoi(optarg);
		break;
	    }

	    case 's':
	    {
		type = atoi(optarg);
		if (type == S6000)
		{
		    reader = &s6000_info;
		    break;
		}

		if (type == S6500)
		{
		    reader = &s6500_info;
		    break;
		}

		fprintf(stderr, "Unsupported tag reader model: %d\n", type);
		usage(argc, argv);
		exit(1);
	    }

	    case 'w':
	    {
		wood_hack = 1;
		break;
	    }

	    case 'h':
	    {
		usage(argc, argv);
		exit(0);
	    }

	    case 'V':
	    {
		printf(PACKAGE " version " VERSION "\n");
		exit(0);
	    }

	    case 'v':
	    {
		verbosity++;
		break;
	    }

	    case '?':
	    {
		usage(argc, argv);
		exit(1);
	    }

	    default:
	    {
		abort();
	    }
	}
    }

    /* Make sure we have at least one source defined */
    if (antenna_count == 0)
    {
        sources[antenna_count++] = "unknown source";
    }
    
    /* The next args is the name of the tag file */
    if (! (optind < argc))
    {
	usage(argc, argv);
	exit(1);
    }

    /* Open the file containing the tag ids and names */
    if ((stream = fopen(argv[optind], "r")) == NULL)
    {
	perror("Unable to open tag-file");
	exit(1);
    }

    /* Parse it */
    if (read_ids(stream, &tags, &tag_count) < 0)
    {
	exit(1);
    }

    /* Make sure we have at least one tag */
    if (tag_count == 0)
    {
	fprintf(stderr, "No tags found in %s\n", argv[optind]);
    }

    /* Clean up */
    fclose(stream);

    if (0 < verbosity)
    {
	printf("read %d tags from %s\n", tag_count, argv[optind]);
    }

    optind++;

    /* No more args allowed */
    if (optind < argc)
    {
	usage(argc, argv);
	exit(1);
    }

    /* Initialize the counts to zero */
    memset(current_id_count, 0, sizeof(int) * MAX_ANTENNAS);

#ifdef WIN32
    /* Open the serial port */
    port = init_serial_port(device, type);
#else /* ! WIN32 */
    /* Open the serial port */
    fd = init_serial_port(device, type);
#endif

#if defined(WIN32) && defined(KEYBOARD_CONTROL)
    /* Open the console */
    if ((console = GetStdHandle(STD_INPUT_HANDLE)) == INVALID_HANDLE_VALUE)
    {
	perror("GetStdHandle() failed");
	exit(1);
    }
#endif

    /* Initiate an attempt to connect to a server */
    fprintf(stderr, "Connecting to the elvin server..."); fflush(stderr);
    if (! elvin_async_connect(handle, connect_cb, NULL, error)) {
	fprintf(stderr, "  failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Go into the main loop */
    if (! elvin_mainloop_run(client, error)) {
	fprintf(stderr, "elvin_mainloop_run() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Close the connection */
    if (! elvin_sync_disconnect(handle, error)) {
	fprintf(stderr, "elvin_sync_disconnect() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Free the handle */
    if (! elvin_handle_free(handle, error)) {
	fprintf(stderr, "elvin_handle_free() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Clean up the sync API */
    if (! elvin_sync_cleanup(1, error)) {
	fprintf(stderr, "elvin_sync_cleanup() failed\n");
	elvin_error_fprintf(stderr, error);
	exit(1);
    }

    /* Free the error context */
    if (! elvin_error_free(error, NULL)) {
	fprintf(stderr, "elvin_error_free() failed\n");
	exit(1);
    }

    exit(0);
}
